from copy import deepcopy
import pytest
import numpy as np
from matplotlib import pyplot as plt

from ctapipe.image.cleaning import tailcuts_clean
from ctapipe.image.geometry_converter import convert_geometry_1d_to_2d, \
    convert_geometry_back
from ctapipe.image.hillas import hillas_parameters, HillasParameterizationError
from ctapipe.instrument import CameraGeometry
from ctapipe.io.hessio import hessio_event_source
from ctapipe.utils import get_dataset
from ctapipe.visualization import CameraDisplay
from ctapipe.image.toymodel import generate_2d_shower_model, make_toymodel_shower_image
from astropy import units as u

#cam_ids = CameraGeometry.get_known_camera_names()
cam_ids = ['FlashCam',]

@pytest.mark.parametrize("rot", [3,])
@pytest.mark.parametrize("cam_id", cam_ids)
def test_convert_geometry(cam_id, rot):

    geom = CameraGeometry.from_name(cam_id)

    if geom.pix_type=='rectangular':
        return  # skip non-hexagonal cameras, since they don't need conversion

    model = generate_2d_shower_model(centroid=(0.4, 0), width=0.01, length=0.03,
                                     psi="25d")

    _,image,_ = make_toymodel_shower_image(geom, model.pdf,
                                           intensity=50,
                                           nsb_level_pe=100)

    hillas_0 = hillas_parameters(geom.pix_x, geom.pix_y, image)

    geom2d, image2d = convert_geometry_1d_to_2d(geom, image,
                                                 geom.cam_id+str(rot),
                                                 add_rot=-2)
    geom1d, image1d = convert_geometry_back(geom2d, image2d,
                                            geom.cam_id+str(rot),
                                            add_rot=rot)
    hillas_1 = hillas_parameters(geom1d.pix_x, geom1d.pix_y, image1d)

    if __name__ == "__main__":
        plot_cam(geom, geom2d, geom1d, image, image2d, image1d)

    assert np.abs(hillas_1.phi - hillas_0.phi).deg < 1.0
    # TODO: test other parameters

def plot_cam(geom, geom2d, geom1d, image, image2d, image1d):
    plt.viridis()
    plt.figure(figsize=(12, 3))
    ax = plt.subplot(1, 3, 1)
    CameraDisplay(geom, image=image)
    plt.subplot(1, 3, 2, sharex=ax, sharey=ax)
    CameraDisplay(geom2d, image=image2d)
    plt.subplot(1, 3, 3, sharex=ax, sharey=ax)
    CameraDisplay(geom1d, image=image1d)


if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.DEBUG)
    test_convert_geometry("FlashCam", 3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# coding: utf8

class ConsumerSequential():

    """`ConsumerSequential` class represents a Consumer pipeline Step.
    """

    def __init__(
            self, coroutine, name=None):
        """
        Parameters
        ----------
        coroutine : Class instance that contains init, run and finish methods
        """
        self.name = name
        self.coroutine = coroutine
        self.running = 0
        self.nb_job_done = 0


    def init(self):
        """
        Initialise coroutine sockets and poller
        Returns
        -------
        True if coroutine init method returns True, otherwise False
        """
        if self.name is None:
            self.name = "Consumer"
        if self.coroutine is None:
            return False
        if self.coroutine.init() == False:
            return False
        return True

    def run(self,inputs=None):
        """ Executes coroutine run method

        Parameters
        ----------
        inputs: input for coroutine.run
        """
        self.coroutine.run(inputs)
        self.nb_job_done+=1

    def finish(self):
        """
        Call coroutine finish method
        """
        self.coroutine.finish()
        return True



from os.path import join, dirname
from ctapipe.utils import get_dataset
from ctapipe.io.eventfilereader import EventFileReader, \
    EventFileReaderFactory, HessioFileReader


def test_event_file_reader():
    try:
        EventFileReader(config=None, tool=None)
    except TypeError:
        return
    raise TypeError("EventFileReader should raise a TypeError when "
                    "instantiated due to its abstract methods")


def test_hessio_file_reader():
    dataset = get_dataset("gamma_test.simtel.gz")
    file = HessioFileReader(None, None, input_path=dataset)
    assert file.directory == dirname(dataset)
    assert file.extension == ".gz"
    assert file.filename == "gamma_test.simtel"
    source = file.read()
    event = next(source)
    assert event.r0.tels_with_data == {38, 47}


def test_get_event():
    dataset = get_dataset("gamma_test.simtel.gz")
    file = HessioFileReader(None, None, input_path=dataset)
    event = file.get_event(2)
    assert event.count == 2
    assert event.r0.event_id == 803
    event = file.get_event(803, True)
    assert event.count == 2
    assert event.r0.event_id == 803


def test_get_num_events():
    dataset = get_dataset("gamma_test.simtel.gz")
    file = HessioFileReader(None, None, input_path=dataset)
    num_events = file.num_events
    assert(num_events == 9)

    file.max_events = 2
    num_events = file.num_events
    assert (num_events == 2)


def test_event_file_reader_factory():
    dataset = get_dataset("gamma_test.simtel.gz")
    factory = EventFileReaderFactory(None, None)
    factory.input_path = dataset
    cls = factory.get_class()
    file = cls(None, None)
    num_events = file.num_events
    assert(num_events == 9)



"""
Calibrator for the R0 -> R1 data level transition.

This module handles the calibration from the R0 data level to R1. This data
level transition will be handled by the camera servers, not in the pipeline,
however the pipeline can be used as a test-bench in the comissioning stage of
the cameras.

As the R1 calibration is camera specific, each camera (and seperately the MC)
requires their own calibrator class with inherits from `CameraR1Calibrator`.
`HessioR1Calibrator` is the calibrator for the MC data obtained from readhess.
Through the use of `CameraR1CalibratorFactory`, the correct
`CameraR1Calibrator` can be obtained based on the origin (MC/Camera format)
of the data.
"""
from traitlets import CaselessStrEnum, Unicode
from ctapipe.core import Component, Factory
from abc import abstractmethod

__all__ = ['HessioR1Calibrator', 'CameraR1CalibratorFactory']

CALIB_SCALE = 1.05
"""
CALIB_SCALE is only relevant for MC calibration.

CALIB_SCALE is the factor needed to transform from mean p.e. units to units of
the single-p.e. peak: Depends on the collection efficiency, the asymmetry of
the single p.e. amplitude  distribution and the electronic noise added to the
signals. Default value is for GCT.

To correctly calibrate to number of photoelectron, a fresh SPE calibration
should be applied using a SPE sim_telarray run with an artificial light source.
"""


class CameraR1Calibrator(Component):
    name = 'CameraR1Calibrator'
    origin = None

    def __init__(self, config, tool, **kwargs):
        """
        Parent class for the r1 calibrators. Fills the r1 container.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool or None
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, parent=tool, **kwargs)
        if self.origin is None:
            raise ValueError("Subclass of CameraR1Calibrator should specify "
                             "an origin")

        self._r0_empty_warn = False

    @abstractmethod
    def calibrate(self, event):
        """
        Abstract method to be defined in child class.

        Perform the conversion from raw R0 data to R1 data
        (ADC Samples -> PE Samples), and fill the r1 container.

        Parameters
        ----------
        event : container
            A `ctapipe` event container
        """

    def check_r0_exists(self, event, telid):
        """
        Check that r0 data exists. If it does not, then do not change r1.

        This ensures that if the containers were filled from a file containing
        r0 data, it is not overwritten by non-existant data.

        Parameters
        ----------
        event : container
            A `ctapipe` event container
        telid : int
            The telescope id.

        Returns
        -------
        bool
            True if r0.tel[telid].adc_samples is not None, else false.
        """
        r0 = event.r0.tel[telid].adc_samples
        if r0 is not None:
            return True
        else:
            if not self._r0_empty_warn:
                self.log.warning("Encountered an event with no R0 data. "
                                 "R1 is unchanged in this circumstance.")
                self._r0_empty_warn = True
            return False


class HessioR1Calibrator(CameraR1Calibrator):
    name = 'HessioR1Calibrator'
    origin = 'hessio'

    def calibrate(self, event):
        if event.meta['origin'] != 'hessio':
            raise ValueError('Using HessioR1Calibrator to calibrate a '
                             'non-hessio event.')

        for telid in event.r0.tels_with_data:
            if self.check_r0_exists(event, telid):
                try:
                    samples = event.r0.tel[telid].adc_samples
                    n_samples = samples.shape[2]
                except IndexError:
                    # To handle ASTRI
                    samples = event.r0.tel[telid].adc_sums[..., None]
                    n_samples = samples.shape[2]
                ped = event.mc.tel[telid].pedestal / n_samples
                gain = event.mc.tel[telid].dc_to_pe * CALIB_SCALE
                calibrated = (samples - ped[..., None]) * gain[..., None]
                event.r1.tel[telid].pe_samples = calibrated


# External Children
try:
    from targetpipe.calib.camera.r1 import TargetioR1Calibrator
except ImportError:
    pass


class CameraR1CalibratorFactory(Factory):
    name = "CameraR1CalibratorFactory"
    description = "Obtain CameraR1Calibrator based on file origin"

    subclasses = Factory.child_subclasses(CameraR1Calibrator)
    subclass_names = [c.origin for c in subclasses]

    origin = CaselessStrEnum(subclass_names, 'hessio',
                             help='Origin of events to be '
                                  'calibration.').tag(config=True)

    # Product classes traits
    pedestal_path = Unicode('', allow_none=True,
                            help='Path to a pedestal file').tag(config=True)
    tf_path = Unicode('', allow_none=True,
                      help='Path to a Transfer Function file').tag(config=True)
    adc2pe_path = Unicode('', allow_none=True,
                          help='Path to an adc2pe file').tag(config=True)

    def get_factory_name(self):
        return self.name

    def get_product_name(self):
        return self.origin



# skip these tests if matplotlib can't be imported
import pytest
plt = pytest.importorskip("matplotlib.pyplot")

from ctapipe.instrument import CameraGeometry
from ctapipe.io import get_array_layout
from numpy import ones

def test_camera_display_single():
    """ test CameraDisplay functionality """
    from ..mpl import CameraDisplay

    geom = CameraGeometry.from_name("LSTCam")
    disp = CameraDisplay(geom)
    image = ones(len(geom.pix_x), dtype=float)
    disp.image = image
    disp.add_colorbar()
    disp.cmap = 'spectral'
    disp.set_limits_minmax(0, 10)
    disp.set_limits_percent(95)


def test_camera_display_multiple():
    """ create a figure with 2 subplots, each with a CameraDisplay """
    from ..mpl import CameraDisplay
    
    geom = CameraGeometry.from_name("LSTCam")
    fig, ax = plt.subplots(2, 1)

    d1 = CameraDisplay(geom, ax=ax[0])
    d2 = CameraDisplay(geom, ax=ax[1])

    image = ones(len(geom.pix_x), dtype=float)
    d1.image = image
    d2.image = image


def test_array_display():
    from ..mpl import ArrayDisplay

    # load some test data
    layout = get_array_layout("hess")
    X = layout['POSX']
    Y = layout['POSY']
    A = layout['MIRAREA']
    A[:] = 132

    ad = ArrayDisplay(X, Y, A, title="HESS")
    ad.intensities = ones(len(X))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from ctapipe.core import Component
from ctapipe.flow.multiprocess.connections import Connections
from multiprocessing import Process
from multiprocessing import Value
from types import GeneratorType
import zmq

class ProducerZmq(Process, Component, Connections):
    """`ProducerZmq` class represents a Producer pipeline Step.
    It is derived from Process class.
    It gets a Python generator from its coroutine run method.
    It loops overs its generator and sends input to its next stage,
    thanks to its ZMQ REQ socket,
    The process is launched by calling run method.
    init() method is call by run method.
    """
    def __init__(self, coroutine, name, main_connection_name,
                 connections=None):
        """
        Parameters
        ----------
        coroutine : Class instance
            It contains init, run and finish methods
        name: str
            Producer name
        main_connection_name : str
            Default next step name. Used to send data when destination is not provided
        connections: dict {'STEP_NANE' : (zmq STEP_NANE port in)}
            Port number for socket for each next steps
        """
        Process.__init__(self)
        Component.__init__(self,parent=None)
        self.name = name
        Connections.__init__(self, main_connection_name, connections)
        self.coroutine = coroutine
        self.other_requests=dict()
        self._nb_job_done = Value('i',0)
        self._running = Value('i',0)
        self.done = False

    def init(self):
        """
        Initialise coroutine
        Returns
        -------
        True if coroutine init method returns True, otherwise False
        """
        if self.coroutine is None:
            return False
        if self.coroutine.init() == False:
            return False
        self.coroutine.connections = list(self.connections)
        return self.init_connections()

    def run(self):
        """
        Method representing the process’s activity.
        It gets a Python generator from its coroutine run method.
        It loops overs its generator and sends new input to its next stage,
        thanks to its ZMQ REQ socket.
        """
        if self.init() :
            generator = self.coroutine.run()
            if isinstance(generator,GeneratorType):
                for result in generator:
                    self.running = 1
                    self.nb_job_done += 1
                    if isinstance(result,tuple):
                        msg,destination = self.get_destination_msg_from_result(result)
                        self.send_msg(msg,destination)
                    else:
                        self.send_msg(result)
                self.running = 0
            else:
                self.log.warning("Warning: Productor run method was not a python generator.")
        self.finish()
        self.done = True

    def finish(self):
        """
        Executes coroutine method
        """
        self.coroutine.finish()
        return True

    def init_connections(self):
        """
        Initialise zmq sockets.
        Because this class is s Process, This method must be call in the run
         method to be hold by the correct process.
        """
        self.context = zmq.Context()
        Connections.init_connections(self)
        return True

    @property
    def nb_job_done(self):
        return self._nb_job_done.value

    @nb_job_done.setter
    def nb_job_done(self, value):
        self._nb_job_done.value = value

    @property
    def running(self):
        return self._running.value

    @running.setter
    def running(self, value):
        self._running.value = value



import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt

from ctapipe import visualization
from ctapipe.plotting.array import ArrayPlotter, NominalPlotter
from numpy import ceil, sqrt
from ctapipe.core import Component
from ctapipe.core.traits import Bool


class EventViewer(Component):
    """
    Event viewer class built on top of the other plotters to allow a single
    view of both the camera images and the projected Hillas parameters for a
    single event. Can be further modified to show the reconstructed shower
    direction and core position if needed. Plus further info
    """
    name = 'EventViewer'
    test = Bool(True, help='').tag(config=True)

    def __init__(self, draw_hillas_planes=False):
        """

        Parameters
        ----------
        draw_hillas_planes: bool
            Determines whether a projection of the Hillas parameters in the nominal and tilted systems should be drawn
        """
        self.array_view = None
        self.nominal_view = None

        self.cam_display = dict()
        self.draw_hillas_planes = draw_hillas_planes

    def draw_source(self, source):
        """
        Loop over events and draw each

        Parameters
        ----------
        source: ctapipe source object

        Returns
        -------
            None
        """
        for event in source:
            self.draw_event(event)

        return

    def draw_event(self, event, hillas_parameters=None, tilted_system=None):
        """
        Draw display for a given event

        Parameters
        ----------
        event: ctapipe event object
        hillas_parameters: dict
            Dictionary of Hillas parameters (in nominal system)

        Returns
        -------
            None
        """
        tel_list = event.r0.tels_with_data
        images = event.dl1

        # First close any plots that already exist
        plt.close()
        ntels = len(tel_list)

        fig = plt.figure(figsize=(20, 20 * 0.66))

        # If we want to draw the Hillas parameters in different planes we need to split our figure
        if self.draw_hillas_planes:
            y_axis_split = 2
        else:
            y_axis_split = 1

        outer_grid = gridspec.GridSpec(1, y_axis_split, width_ratios=[y_axis_split, 1])
        # Create a square grid for camera images
        nn = int(ceil(sqrt(ntels)))
        nx = nn
        ny = nn

        while nx * ny >= ntels:
            ny-=1
        ny+=1
        while nx * ny >= ntels:
            nx -= 1
        nx += 1

        camera_grid = gridspec.GridSpecFromSubplotSpec(ny, nx, subplot_spec=outer_grid[0])

        # Loop over camera images of all telescopes and create plots
        for ii, tel_id in zip(range(ntels), tel_list):

            ax = plt.subplot(camera_grid[ii])
            geom = event.inst.subarray.tel[tel_id].camera
            self.get_camera_view(tel_id,
                                 image=images.tel[tel_id].image[0],
                                 ax=ax,
                                 geom=geom)

        # If we want to draw the Hillas parameters in different planes we need to make a couple more viewers
        if self.draw_hillas_planes:
            # Split the second sub figure into two further figures
            reco_grid = gridspec.GridSpecFromSubplotSpec(2, 1, subplot_spec=outer_grid[1])
            # Create plot of telescope positions at ground level


            # Draw MC position (this should change later)
            array.draw_position(event.mc.core_x, event.mc.core_y, use_centre=True)
            array.draw_array(((-300,300),(-300,300)))

            # If we have valid Hillas parameters we should draw them in the Nominal system
            if hillas_parameters is not None:
                array.overlay_hillas(hillas_parameters, draw_axes=True)

                nominal =  NominalPlotter(hillas_parameters=hillas_parameters, draw_axes=True, ax=plt.subplot(reco_grid[1]))
                nominal.draw_array()
        self.nominal_view = nominal
        self.array_view = array

        plt.show()

        return

    def get_camera_view(self, tel_id, image, axis, geom):
        """
        Create camera viewer for a given camera image

        Parameters
        ----------
        tel_id: int
            Telescope ID number
        image: ndarray
            Array of calibrated pixel intensities
        axis: matplotlib axis
            Axis on which to draw plot
        geom: ctapipe.instrument.CameraGeometry
            Camera geometry for this tel_id

        Returns
        -------
            Camera display
        """
        #if tel_id not in self.cam_display:
        # Argh this is annoying, for some reason we cannot cahe the displays
        self.cam_display[tel_id] = visualization.CameraDisplay(geom, title="CT{"
                                                                        "0}".format(tel_id))
        self.cam_display[tel_id].add_colorbar()
        self.cam_display[tel_id].pixels.set_antialiaseds(False)
        self.cam_display[tel_id].autoupdate = True
        self.cam_display[tel_id].cmap = "viridis"

        self.cam_display[tel_id].ax = axis
        self.cam_display[tel_id].image = image

        self.cam_display[tel_id].set_limits_percent(95)

        return self.cam_display[tel_id]


"""
This file is deprecated, but parts may be used to load other instrumental
descriptions from MC config files.
"""

import os
import random
import textwrap

import numpy as np
from astropy import units as u
from astropy.table import Table

from ctapipe.io.files import get_file_type

__all__ = ['load', 'load_fakedata', 'load_config',
           'read_config_data', 'get_var_from_file', 'write_fits']


def load(filename='', path=None, version='', instr_item='', telID=''):
    """
    Function to load instrument data from a file
    
    Parameters
    ----------
    filename: string
        name of the file
    path: string
        name of the path, where the file(s) are located
        - when already given as part of the filename, nothing to be entered
        - when file(s) is located in current directory, nothing to be entered
    version: string
        needed if no filename is given and if files with a given version should
        be looked for
    instr_item: string
        item of the instrument (Telescope,Camera or Optics) whose data
        should be loaded, can be either a list, e.g. ('Camera','Telescope') or
        just one string, e.g. 'Camera'

    Returns
    -------
    possibility 0:
    telescope,camera,optics = load_fakedata(filename)
    
    possibility 1:
    telescope,camera,optics = load_hessio(filename)
    
    possibility 2:
    telescope,camera,optics = load_fits(filename = '',path = None,
                                        version = '',instr_item = '')
    possibility 3:
    telescope,camera,optics = load_config(filename)
    
    all dictionaries contain astropy.table Tables
    """
    filetype = get_file_type(filename)

    if filename == '' and version == '':
        return load_fakedata()
    elif filetype == 'cfg':
        return load_config(filename)
    else:
        raise TypeError("File type {} not supported".format(filetype))


def load_fakedata():
    """
    Function writing faked date into an astropy.table Table
    
    Returns
    -------
    telescope,camera,optics: 3 dictionaries
        all dictionaries contain astropy.table Tables
    """
    print('Faked data will be produced.')
    version = 'Feb2016'
    tel_num = 10
    tel_id = [random.randrange(1, 124, 1) for x in range(0, tel_num)]

    telescope = {}
    camera = {}
    optics = {}

    tel_table_prime = Table()
    tel_table_prime['TelID'] = tel_id

    tel_posX = [random.uniform(1, 100) for x in range(tel_num)]
    tel_posY = [random.uniform(1, 100) for x in range(tel_num)]
    tel_posZ = [random.uniform(1, 100) for x in range(tel_num)]
    tel_table_prime['TelX'] = tel_posX
    tel_table_prime['TelX'].unit = u.m
    tel_table_prime['TelY'] = tel_posY
    tel_table_prime['TelY'].unit = u.m
    tel_table_prime['TelZ'] = tel_posZ
    tel_table_prime['TelZ'].unit = u.m
    mirror_area = [random.uniform(1, 100) for x in range(tel_num)]
    tel_table_prime['MirA'] = mirror_area
    tel_table_prime['MirA'].unit = u.m ** 2
    mirror_num = [random.randrange(1, 124, 1) for x in range(tel_num)]
    tel_table_prime['MirN'] = mirror_num
    foclen = [random.uniform(1, 100) for x in range(tel_num)]
    tel_table_prime['FL'] = foclen
    tel_table_prime['FL'].unit = u.m

    telescope['TelescopeTable_Version%s' % version] = tel_table_prime

    for t in range(len(tel_id)):
        cam_table_prime = Table()
        opt_table_prime = Table()
        pixel_num = 128
        pix_posX = [random.uniform(1, 100) for x in range(tel_num * pixel_num)]
        pix_posY = [random.uniform(1, 100) for x in range(tel_num * pixel_num)]
        pix_id = np.arange(len(pix_posX))
        pix_area = [random.uniform(1, 100) for x in range(tel_num * pixel_num)]

        cam_table_prime.meta = {'TELID': tel_id[t], 'VERSION': version, \
                                'PIXX_DES': 'x-position of the pixel measured '
                                            'by...'}
        cam_table_prime['PixID'] = pix_id
        cam_table_prime['PixX'] = pix_posX
        cam_table_prime['PixX'].unit = u.m
        cam_table_prime['PixY'] = pix_posY
        cam_table_prime['PixY'].unit = u.m
        cam_table_prime['PixA'] = pix_area
        cam_table_prime['PixA'].unit = u.mm ** 2

        opt_table_prime.meta = {'TELID': tel_id[t], 'VERSION': version, \
                                'MIRN': mirror_num[t],
                                'MIRA': mirror_area[t] * u.m ** 2, \
                                'FL': foclen[t] * u.m, \
                                'MIRA_DES': 'Area of all mirrors'}
        tab_mirrefl = Table()
        tab_mirrefl['wavel'] = np.arange(100, 700, 10)
        tab_mirrefl['wavel'].unit = u.nm
        tab_mirrefl['refl'] = [random.uniform(0.01, 1) \
                               for x in range(len(tab_mirrefl['wavel']))]

        opt_table_prime['MirR'] = tab_mirrefl

        camera['CameraTable_Version%s_TelID%i' % (version, tel_id[t])] \
            = cam_table_prime
        optics['OpticsTable_Version%s_TelID%i' % (version, tel_id[t])] \
            = opt_table_prime
    print('Astropy tables have been created.')
    return telescope, camera, optics




def load_config(filename):
    """
    Function to load config file data into a table
    
    Parameters
    ----------
    filename: string
        name of the file

    Returns
    -------
    telescope,camera,optics: 3 dictionaries
        all dictionaries contain astropy.table Tables
    """
    data = read_config_data(filename)
    dirname = os.path.dirname(filename)
    version = os.path.splitext(os.path.basename(filename))[0]

    # Creating 3 dictionaries where the instrument configuration will be stored
    # The dictionaries themselves contain astropy.table.Table objects
    telescope = {}
    camera = {}
    optics = {}

    # --------------------------------------------------------------------------
    # Telescope configuration
    tel_table_prime = Table()
    tel_table_prime['TelID'] = [1, 2]
    tel_table_prime.meta = {'VERSION': version}
    try:
        tel_table_prime['TriggerDelayComp'] = \
            data.trigger_delay_compensation[0]
    except:
        pass
    try:
        tel_table_prime['DefaultTrigger'] = data.default_trigger[0]
    except:
        pass
    try:
        tel_table_prime['TriggerDiscBins'] = data.disc_bins[0]
    except:
        pass
    try:
        tel_table_prime['TriggerDiscStart'] = data.disc_start[0]
    except:
        pass
    try:
        tel_table_prime['DiscriminatorPulseShape'] = \
            data.discriminator_pulse_shape[0]
    except:
        pass
    try:
        pulseshape_table = Table()
        pulseshape_filename = \
            dirname + '/' + textwrap.dedent(
                data.discriminator_pulse_shape[0]).strip()
        pulseshape_version = \
            os.path.splitext(os.path.basename(pulseshape_filename))[0]
        time, amplitude = \
            np.loadtxt(pulseshape_filename, unpack=True, usecols=(0, 1))
        pulseshape_table.meta = {'NAME': 'Discriminator pulse shape',
                                 'VERSION': pulseshape_version}
        pulseshape_table['Time'] = time
        pulseshape_table['Time'].unit = u.ns
        pulseshape_table['Amplitude'] = amplitude
        telescope['Tel_DiscriminatorPulseShape'] = pulseshape_table
    except:
        pass
    try:
        tel_table_prime['DiscriminatorAmplitude'] = \
            data.discriminator_amplitude[0]
        tel_table_prime['DiscriminatorAmplitude'].unit = u.mV
    except:
        pass
    try:
        tel_table_prime['NumTriggerPixels'] = trigger_pixels[0]
    except:
        pass
    try:
        tel_table_prime['DiscriminatorThreshold'] = \
            discriminator_threshold[0]
    except:
        pass
    try:
        tel_table_prime['DefaultTrig'] = data.default_trigger[0]
    except:
        pass
    try:
        tel_table_prime['TelTrigMinTime'] = data.teltrig_min_time[0]
        tel_table_prime['TelTrigMinTime'].unit = u.ns
    except:
        pass
    try:
        tel_table_prime['TelTrigMinSigSum'] = data.teltrig_min_sigsum[0]
        tel_table_prime['TelTrigMinSigSum'].unit = u.pV * u.s
    except:
        pass
    try:
        tel_table_prime['TrigDelayComp'] = data.trigger_delay_compensation[0]
    except:
        pass
    try:
        tel_table_prime['DiscriminatorVarThreshold'] = \
            discriminator_var_threshold[0]
    except:
        pass
    try:
        tel_table_prime['SamplingRate'] = data.fadc_MHz[0]
        tel_table_prime['SamplingRate'].unit = u.MHz
    except:
        pass
    try:
        tel_table_prime['SamplingRate'] = data.fadc_mhz[0]
        tel_table_prime['SamplingRate'].unit = u.MHz
    except:
        pass
    try:
        tel_table_prime['FADCPulseShape'] = data.fadc_pulse_shape[0]
    except:
        pass
    try:
        fadc_pulseshape_table = Table()
        fadc_pulseshape_filename = \
            dirname + '/' + textwrap.dedent(data.fadc_pulse_shape[0]).strip()
        pulseshape_version = \
            os.path.splitext(os.path.basename(fadc_pulseshape_filename))[0]
        time, amplitude = \
            np.loadtxt(fadc_pulseshape_filename, unpack=True, usecols=(0, 1))
        fadc_pulseshape_table.meta = {'NAME': 'FADC pulse shape',
                                      'VERSION': pulseshape_version}
        fadc_pulseshape_table['Time'] = time
        fadc_pulseshape_table['Time'].unit = u.ns
        fadc_pulseshape_table['Amplitude'] = amplitude
        telescope['Tel_FADCPulseShape'] = pulseshape_table
    except:
        pass
    try:
        tel_table_prime['FADCBins'] = data.fadc_bins[0]
    except:
        pass
    try:
        tel_table_prime['FADCSumBins'] = data.fadc_sum_bins[0]
    except:
        pass
    try:
        tel_table_prime['FADCSumOffset'] = data.fadc_sum_offset[0]
    except:
        pass
    try:
        tel_table_prime['FADCPedestal'] = data.fadc_pedestal[0]
    except:
        pass
    try:
        tel_table_prime['FADCAmplitude'] = data.fadc_amplitude[0]
    except:
        pass
    try:
        tel_table_prime['FADCNoise'] = data.fadc_noise[0]
    except:
        pass
    try:
        tel_table_prime['NumGains'] = data.num_gains[0]
    except:
        pass
    try:
        tel_table_prime['FADCLowGainPedestal'] = data.fadc_lg_pedestal[0]
    except:
        pass
    try:
        tel_table_prime['FADCLowGainAmplitude'] = data.fadc_lg_amplitude[0]
    except:
        pass
    try:
        tel_table_prime['FADCLowGainNoise'] = data.fadc_lg_noise[0]
    except:
        pass
    try:
        tel_table_prime['FADCMaxSignal'] = data.fadc_max_signal[0]
    except:
        pass
    try:
        tel_table_prime['FADCMaxSum'] = data.fadc_max_sum[0]
    except:
        pass
    try:
        tel_table_prime['CamBodyDiam'] = data.camera_body_diameter[0]
        tel_table_prime['CamBodyDiam'].unit = u.cm
    except:
        pass
    try:
        tel_table_prime['CamDepth'] = data.camera_depth[0]
        tel_table_prime['CamDepth'].unit = u.cm
    except:
        pass
    try:
        tel_table_prime['CamConfig'] = data.camera_config_file[0]
    except:
        pass
    try:
        tel_table_prime['PixNum'] = data.camera_pixels[0]
    except:
        pass

    # Beside other tables containimng telescope configuration data, the main
    # telescope table is written into the telescope dictionary.
    telescope['TelescopeTable_%s' % version] = tel_table_prime

    # --------------------------------------------------------------------------
    # Camera configuration
    cam_table_prime = Table()
    cam_table_prime.meta = {'VERSION': version}
    try:
        cam_table_prime.meta['CamBodyDiam'] = \
            data.camera_body_diameter[0] * u.cm
    except:
        pass
    try:
        cam_table_prime.meta['CamDepth'] = data.camera_depth[0] * u.cm
    except:
        pass
    try:
        cam_table_prime.meta['CamConfig'] = data.camera_config_file[0]
    except:
        pass
    try:
        cam_config_filename = \
            dirname + '/' + textwrap.dedent(data.camera_config_file[0]).strip()
        pix_posX, pix_posY = \
            np.loadtxt(cam_config_filename,
                       comments=(
                       '#', 'M', 'PixType', 'AnalogSumTrigger', 'Rotate', \
                       'Trigger', 'DigitalSumTrigger'),
                       usecols=(3, 4), unpack=True)
        board_ID = \
            np.loadtxt(cam_config_filename,
                       comments=(
                       '#', 'M', 'PixType', 'AnalogSumTrigger', 'Rotate', \
                       'Trigger', 'DigitalSumTrigger'),
                       usecols=(7,), unpack=True, dtype=str)
        try:
            pix_ID, module_number, board_number, channel_number, pixel_on = \
                np.loadtxt(cam_config_filename,
                           comments=(
                           '#', 'M', 'PixType', 'AnalogSumTrigger', 'Rotate', \
                           'Trigger', 'DigitalSumTrigger'),
                           usecols=(1, 2, 5, 6, 8), unpack=True, dtype=int)

        except:
            pix_ID, module_number, board_number, channel_number = \
                np.loadtxt(cam_config_filename,
                           comments=(
                           '#', 'M', 'PixType', 'AnalogSumTrigger', 'Rotate', \
                           'Trigger', 'DigitalSumTrigger'),
                           usecols=(1, 2, 5, 6), unpack=True, dtype=int)
            pixel_on = [-1]
        cam_table_prime['PixID'] = pix_ID
        cam_table_prime['PixX'] = pix_posX
        cam_table_prime['PixX'].unit = u.cm
        cam_table_prime['PixY'] = pix_posY
        cam_table_prime['PixY'].unit = u.cm
        cam_table_prime['ModuleNum'] = module_number
        cam_table_prime['BoardNum'] = board_number
        cam_table_prime['ChannelNum'] = channel_number
        cam_table_prime['BoardID'] = board_ID
        if pixel_on[0] != -1:
            cam_table_prime['PixelON'] = pixel_on
    except:
        cam_table_prime['PixID'] = [1, 2]
    try:
        cam_config_filename = \
            dirname + '/' + textwrap.dedent(data.camera_config_file[0]).strip()
        try:
            pix_type, pmt_type, cathode_shape_type, visible_cathode_diameter, \
            funnel_shape_type, funnel_diameter, funnel_depth = \
                np.genfromtxt(cam_config_filename,
                              usecols=(1, 2, 3, 4, 5, 6, 7), \
                              unpack=True, max_rows=1)
            funnel_efficiency, funnel_wall_reflectivity = \
                np.genfromtxt(cam_config_filename, usecols=(8, 9),
                              unpack=True, \
                              max_rows=1, dtype=None)
        except:
            pix_type, pmt_type, cathode_shape_type, visible_cathode_diameter, \
            funnel_shape_type, funnel_diameter, funnel_depth = \
                np.genfromtxt(cam_config_filename,
                              usecols=(1, 2, 3, 4, 5, 6, 7), \
                              unpack=True, max_rows=1)
            funnel_efficiency = \
                np.genfromtxt(cam_config_filename, usecols=(8,), \
                              unpack=True, max_rows=1, dtype=str)
            funnel_wall_reflectivity = [-1]
        cam_table_prime['PixType'] = pix_type
        cam_table_prime['PMTType'] = pmt_type
        cam_table_prime['CathType'] = cathode_shape_type
        cam_table_prime['CathDiam'] = visible_cathode_diameter
        cam_table_prime['CathDiam'].unit = u.cm
        cam_table_prime['FunnelTzpe'] = funnel_shape_type
        cam_table_prime['FunnelDiam'] = funnel_diameter
        cam_table_prime['FunnelDiam'].unit = u.cm
        cam_table_prime['FunnelDepth'] = funnel_depth
        cam_table_prime['FunnelDepth'].unit = u.cm
        cam_table_prime['FunnelEff'] = funnel_efficiency.reshape(1)[0]
        if funnel_wall_reflectivity[0] != -1:
            cam_table_prime['FunnelRefl'] = funnel_wall_reflectivity
    except:
        pass
    try:
        funnel_eff_table = Table()
        funnel_eff_filename = \
            dirname + '/' + funnel_efficiency.reshape(1)[0]
        funnel_eff_version = \
            os.path.splitext(os.path.basename(funnel_eff_filename))[0]
        angle, transmission = \
            np.loadtxt(funnel_eff_filename, unpack=True)
        refl_table.meta = {'NAME': 'Funnel efficiency',
                           'VERSION': funnel_eff_version}
        funnel_eff_table['Angle'] = wavel
        funnel_eff_table['Angle'].unit = u.degree
        funnel_eff_table['Transmission'] = transmission
        camera['Cam_FunnelEfficiency'] = funnel_eff_table
    except:
        pass
    try:
        cam_table_prime.meta['PixNum'] = data.camera_pixels[0]
    except:
        pass
    try:
        cam_table_prime['PhDetEff'] = data.quantum_efficiency[0]
    except:
        pass
    try:
        quantumeff_table = Table()
        cam_quantumeff_filename = \
            dirname + '/' + textwrap.dedent(data.quantum_efficiency[0]).strip()
        cam_quantumeff_version = \
            os.path.splitext(os.path.basename(cam_quantumeff_filename))[0]
        wavel, ph_detection_eff = \
            np.loadtxt(cam_quantumeff_filename, unpack=True)
        quantumeff_table.meta = {'NAME': 'Quantum efficiency',
                                 'VERSION': cam_quantumeff_version}
        quantumeff_table['Wavelength'] = wavel
        quantumeff_table['Wavelength'].unit = u.nm
        quantumeff_table['DetEfficiency'] = ph_detection_eff
        camera['Cam_PhotonDetectionEfficiency'] = quantumeff_table
    except:
        pass
    try:
        cam_table_prime['PESpec'] = data.pm_photoelectron_spectrum[0]
    except:
        pass
    try:
        pespectrum_table = Table()
        pespectrum_filename = \
            dirname + '/' + textwrap.dedent(
                data.pm_photoelectron_spectrum[0]).strip()
        pespectrum_version = \
            os.path.splitext(os.path.basename(pespectrum_filename))[0]
        amplitude, probability = \
            np.loadtxt(pespectrum_filename, unpack=True, usecols=(0, 1))
        pespectrum_table.meta = {'NAME': 'Photoelectron spectrum',
                                 'VERSION': pespectrum_version}
        pespectrum_table['Amplitude [PE]'] = amplitude
        pespectrum_table['Probability'] = probability
        camera['Cam_PhotoelectronSpectrum'] = \
            pespectrum_table
    except:
        pass
    try:
        cam_table_prime['TransTimeJitt'] = data.transit_time_jitter[0]
    except:
        pass
    try:
        cam_table_prime['GainVar'] = data.gain_variation[0]
    except:
        pass
    try:
        cam_table_prime['QEVar'] = data.qe_variation[0]
    except:
        pass
    try:
        cam_table_prime['CamFilter'] = data.camera_filter[0]
    except:
        pass
    try:
        camfilter_table = Table()
        camfilter_filename = \
            dirname + '/' + textwrap.dedent(data.camera_filter[0]).strip()
        camfilter_version = \
            os.path.splitext(os.path.basename(camfilter_filename))[0]
        wavel, transmission = \
            np.loadtxt(camfilter_filename, unpack=True)
        camfilter_table.meta = {'NAME': 'Camera window transmission',
                                'VERSION': camfilter_version}
        camfilter_table['Wavelength'] = wavel
        camfilter_table['Wavelength'].unit = u.nm
        camfilter_table['Transmission'] = transmission
        camera['CameraWindowTransmission'] = camfilter_table
    except:
        pass

    # some variables which are set for simulation --> no real instrument data

    # try: cam_table_prime.meta['MINPH'] = data.min_photons[0]
    # except: pass
    # try: cam_table_prime.meta['MIPE'] = data.min_photoelectrons[0]
    # except: pass
    # try: cam_table_prime.meta['STOREPE'] = data.store_photoelectrons[0]
    # except: pass
    # try: cam_table_prime.meta['NSB'] = data.nightsky_background[0]
    # except: pass

    # Beside other tables containimng camer configuration data, the main
    # camera table is written into the camera dictionary.
    camera['CameraTable_%s' % version] = cam_table_prime

    # --------------------------------------------------------------------------
    # Optics configuration
    opt_table_prime = Table()
    opt_table_prime.meta = {'VERSION': version}

    try:
        opt_table_prime.meta['MirrorList'] = data.mirror_list[0]
    except:
        pass
    try:
        opt_mirrorlist_filename = \
            dirname + '/' + textwrap.dedent(data.mirror_list[0]).strip()
        file = open(opt_mirrorlist_filename, 'r')
        temp_filename = 'mirrorlist_temp.txt'
        temp_file = open(temp_filename, 'w')
        for line in file:
            if '#%' in line:
                line = line.replace('#%', ' ')
            temp_file.write(line)
        file.close()
        temp_file.close()
        mirror_characs = np.loadtxt(temp_filename, unpack=True)
        os.remove(temp_filename)
        try:
            mir_ID = mirror_characs[6]
            opt_table_prime['MirID'] = mir_ID.astype(int)
        except:
            pass
        try:
            mir_posX = mirror_characs[0]
            opt_table_prime['MirX'] = mir_posX
            opt_table_prime['MirX'].unit = u.cm
        except:
            pass
        try:
            mir_posY = mirror_characs[1]
            opt_table_prime['MirY'] = mir_posY
            opt_table_prime['MirY'].unit = u.cm
        except:
            pass
        try:
            mir_posZ = mirror_characs[5]
            opt_table_prime['MirZ'] = mir_posZ
            opt_table_prime['MirZ'].unit = u.cm
        except:
            pass
        try:
            mir_diam = mirror_characs[2]
            opt_table_prime['MirDiam'] = mir_diam
            opt_table_prime['MirDiam'].unit = u.cm
        except:
            pass
        try:
            mir_FL = mirror_characs[3]
            opt_table_prime['MirFL'] = mir_FL
            opt_table_prime['MirFL'].unit = u.cm
        except:
            pass
        try:
            mir_shape_code = mirror_characs[4]
            opt_table_prime['MirShCode'] = mir_shape_code.astype(int)
        except:
            pass
        try:
            mir_nx = mirror_characs[7]
            opt_table_prime['n_x'] = mir_nx
        except:
            pass
        try:
            mir_ny = mirror_characs[8]
            opt_table_prime['n_y'] = mir_ny
        except:
            pass
        try:
            mir_nz = mirror_characs[9]
            opt_table_prime['n_z'] = mir_nz
        except:
            pass
        try:
            dist_optax_mirrorcentrum = mirror_characs[10]
            opt_table_prime['DistOptAxMirCent'] = dist_optax_mirrorcentrum
            opt_table_prime['DistOptAxMirCent'].unit = u.cm
        except:
            pass
    except:
        pass
    try:
        opt_table_prime['MirClass'] = data.mirror_class[0]
    except:
        pass
    try:
        opt_table_prime['PrimMirParam'] = [data.primary_mirror_parameters]
    except:
        pass
    try:
        opt_table_prime['PrimRefRad'] = data.primary_ref_radius[0]
        opt_table_prime['PrimRefRad'].unit = u.cm
    except:
        pass
    try:
        opt_table_prime['PrimDiam'] = data.primary_diameter[0]
        opt_table_prime['PrimDiam'].unit = u.cm
    except:
        pass
    try:
        opt_table_prime['PrimHoleDiam'] = data.primary_hole_diameter[0]
        opt_table_prime['PrimHoleDiam'].unit = u.cm
    except:
        pass
    try:
        opt_table_prime['SndMirParam'] = [data.secondary_mirror_parameters]
    except:
        pass
    try:
        opt_table_prime['SndRefRad'] = data.secondary_ref_radius[0]
        opt_table_prime['SndRefRad'].unit = u.cm
    except:
        pass
    try:
        opt_table_prime['SndDiam'] = data.secondary_diameter[0]
        opt_table_prime['SndDiam'].unit = u.cm
    except:
        pass
    try:
        opt_table_prime['SndHoleDiam'] = data.secondary_hole_diameter[0]
        opt_table_prime['SndHoleDiam'].unit = u.cm
    except:
        pass
    try:
        opt_table_prime['FocalSurfParam'] = [data.focal_surface_param]
    except:
        pass
    try:
        opt_table_prime['FL'] = data.focal_length[0]
        opt_table_prime['FL'].unit = u.cm
    except:
        pass
    try:
        opt_table_prime['FocalSurfRefRad'] = data.focal_surface_ref_radius[0]
        opt_table_prime['FocalSurfRefRad'].unit = u.cm
    except:
        pass
    try:
        opt_table_prime['RandmFocalLen'] = data.random_focal_length[0]
        opt_table_prime['RandmFocalLen'].unit = u.cm
    except:
        pass
    try:
        opt_table_prime['MirReflRndmAngle'] = \
            data.mirror_reflection_random_angle[0]
        opt_table_prime['MirReflRndmAngle'].unit = u.degree
    except:
        pass
    try:
        opt_table_prime['MirAlignRndmDist'] = \
            data.mirror_align_random_distance[0]
        opt_table_prime['MirAlignRndmDist'].unit = u.cm
    except:
        pass
    try:
        opt_table_prime['MirAlignRndmHoriz'] = \
            data.mirror_align_random_horizontal
    except:
        pass
    try:
        opt_table_prime['MirAlignRndmVert'] = \
            data.mirror_align_random_vertical
    except:
        pass
    try:
        opt_table_prime['MirRefl'] = data.mirror_reflectivity[0]
    except:
        pass
    try:
        refl_table = Table()
        mir_refl_filename = \
            dirname + '/' + textwrap.dedent(data.mirror_reflectivity[0]).strip()
        mir_refl_version = \
            os.path.splitext(os.path.basename(mir_refl_filename))[0]
        wavel, reflect = \
            np.loadtxt(mir_refl_filename, unpack=True)
        refl_table.meta = {'NAME': 'Mirror reflectivity',
                           'VERSION': mir_refl_version}
        refl_table['Wavelength'] = wavel
        refl_table['Wavelength'].unit = u.nm
        refl_table['Reflectivity'] = reflect
        optics['Opt_MirrorRefelctivity'] = refl_table
    except:
        pass
    try:
        opt_table_prime['MirList'] = data.mirror_list[0]
    except:
        pass
    try:
        opt_table_prime['MirOffset'] = data.mirror_offset[0]
    except:
        pass
    try:
        opt_table_prime['FocusOffset'] = data.focus_offset[0]
    except:
        pass
    try:
        opt_table_prime['TelTrans'] = data.telescope_transmission[0]
    except:
        pass
    try:
        opt_table_prime['TelRndmAngle'] = data.telescope_random_angle[0]
    except:
        pass
    try:
        opt_table_prime['TelRndmErr'] = data.telescope_random_error[0]
    except:
        pass

    # Beside other tables containimng optics configuration data, the main
    # optics table is written into the optics dictionary.
    optics['OpticsTable_%s' % version] = opt_table_prime

    print('Astropy tables have been created.')
    return (telescope, camera, optics)


def get_var_from_file(filename):
    """
    Function to load and initialize a module implemented as a Python source
    file called `filename` and to return its module objects.
    
    Parameters
    ----------
    filename: ASCII file
        file in which the module objects are defined
    """
    f = open(filename)
    global data
    data = imp.load_source('data', '', f)
    f.close()


def read_config_data(filename):
    """Function to open and load a config file"""
    file = open(filename, 'r')
    print("Config file %s will be opened." % filename)
    temp_filename = \
        '%s_temp.txt' % os.path.splitext(os.path.basename(filename))[0]
    print("Temporary file ", temp_filename, " created.")
    temp_file = open(temp_filename, 'w')
    for line in file:
        if 'echo' in line:
            line = line.replace('echo', '#')
        if "%" in line:
            line = line.replace('%', '#')
        if line.startswith('#'):
            pass
        else:
            if '=' in line:
                index1 = line.index('=')
                if '#' in line:
                    index2 = line.index('#') - 1
                else:
                    index2 = len(line) - 1
                line = line[:index1 + 1] + '[' + line[
                                                 index1 + 1:index2] + ']' + '\n'
                for i in range(index1 + 2, index2 + 1, 1):
                    if line[i] == ' ' or line[i] == 'e' or line[i] == 'E':
                        x = True
                        pass
                    elif ord(line[i]) > 57:
                        x = False
                        break
                    else:
                        x = True
                        pass
                if x == False:
                    line = \
                        line[:index1 + 2] + '"' + line[
                                                  index1 + 2:index2 + 1] + '"' + ']' + '\n'
            else:
                line = '\n'
        line = textwrap.dedent(line)
        temp_file.write(line)
    file.close()
    print("Config file has been closed.")
    temp_file.close()
    print("Temporaryly created file has been closed.")
    get_var_from_file(temp_filename)
    os.remove(temp_filename)
    print("Temporaryly created file has been removed.")
    return data


def write_fits(filename='', path='', instr_dict='', table_name='', \
               overwrite=False):
    """
    Function writing data from the astropy.table Table into a fits file
    Every table is written in an own file; fits files with multiple HDU objects
    cannot be created!
    
    Parameters
    ----------
    filename: string
        if filename is given, the table is written to the given filename
        (path must be part of the filename!), only possible if one specific
        table is given, i.e. table_name and instr_dict have to be given.
        If no filename is given, it is generated automatically using the names
        of the given tables amd instr_dict
    path: string
        pathname of the directory, where the file(s) should be wirtten to;
        default: current directory
    instr_dict: dictionary
        dictionary containing tables with either telescope, camera, or optics
        data; has to specified in every case
    table_name: string
        name of the table of the dictionary which should be written into
        the file; if table_name is given, instr_dict must also be specified
        default: all tables of the dictionary are wirtten to files
    """
    if instr_dict == '':
        return (print('instrument dictionary (instr_dict) must be specified.'))
    if table_name == '':
        for key in instr_dict:
            if filename == '':
                write_name = '%s%s.fits' % (path, key)
                instr_dict[key].write(write_name, overwrite)
                print('%s has been created' % write_name)
            else:
                write_name = filename
                instr_dict[key].write(filename, overwrite)
                print('%s has been created' % write_name)
    else:
        if filename == '':
            write_name = '%s%s.fits' % (path, table_name)
            instr_dict[table_name].write(write_name, overwrite)
            print('%s has been created' % write_name)
        else:
            write_name = filename
            instr_dict[table_name].write(filename, overwrite)
            print('%s has been created' % write_name)

    return (print('Writing process finished.'))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from zmq.error import ZMQError
from zmq import Context
from zmq import SUB
from zmq import Poller
from zmq import POLLIN
from zmq import SUBSCRIBE
from threading import Thread
from pickle import loads
from PyQt4 import QtCore
from PyQt4.QtGui import QLabel
from time import time

class GuiConnexion(Thread, QtCore.QObject):
    """
    Manages communication with pipeline thanks to ZMQ SUB message
    Transmit information to GUI when pipeline change thank to pyqtSignal
    """
    mode_message = QtCore.pyqtSignal(str)
    message = QtCore.pyqtSignal(list)
    reset_message = QtCore.pyqtSignal()

    def __init__(self, gui_port=None, statusBar=None):
        """
        Parameters
        ----------
        gui_port : str
            port to connect for ZMQ communication
        statusBar :  QtGui.QStatusBar
            MainWindow status bar to display information
        """
        Thread.__init__(self)
        QtCore.QObject.__init__(self)
        if gui_port is not None:
            self.context = Context()
            # Socket to talk to pipeline kernel and pipeline steps and router
            self.socket = self.context.socket(SUB)
            gui_adress = "tcp://*:" + str(gui_port)
            try:
                self.socket.bind(gui_adress)
            except ZMQError as e:
                print("ERROR: ".format(str(e), gui_adress))
            # Inform about connection in statusBar
            if statusBar is not None:
                self.statusBar = statusBar
                self.statusBar.insertPermanentWidget(0,QLabel("binded to " + gui_adress))
            # Register socket in a poll and register topics
            self.poll = Poller()
            self.poll.register(self.socket, POLLIN)
            self.socket.setsockopt_string(SUBSCRIBE, 'GUI_GRAPH')
            self.socket.setsockopt_string(SUBSCRIBE, 'GUI_ROUTER_CHANGE')
            self.socket.setsockopt_string(SUBSCRIBE, 'FINISH')
            self.socket.setsockopt_string(SUBSCRIBE, 'MODE')
            # self,stop flag is set by ficnish method to stop this thread
            # properly when GUI is closed
            # self.pipegui will receive new pipeline information
            self.stop = False
            self.steps = list()
            self.config_time = 0
            self.last_send_config = 0
            self.nb_job_done=dict()
        else:
            self.stop = True

    def run(self):
        """
        Method representing the thread’s activity.
        """
        while not self.stop:
            conf_time = time()
            sockets = dict(self.poll.poll(1000))  # Poll or time out (1000ms)
            if self.socket in sockets and sockets[self.socket] == POLLIN:
                # receive a new message form pipeline
                receive = self.socket.recv_multipart()
                topic = receive[0]
                msg = loads(receive[1])
                if topic == b'FINISH':
                    self.flow_has_finish()
                elif topic == b'MODE':
                    self.send_mode(msg)
                else:
                    self.update_full_state(topic,msg)
                    if (conf_time - self.last_send_config) >= 0.0416: # 24 images /sec
                        # inform pipegui
                        self.message.emit(self.steps)
                        self.last_send_config = conf_time
            else:
                if self.steps:
                    self.message.emit(self.steps)
                    self.last_send_config = conf_time

    def reset(self):
        """
        Clear the self.steps list
        """
        self.steps = list()
        self.reset_message.emit()


    def update_full_state(self,topic,msg):
        """
        Redirect topic and message depending on topic

        Parameters
        -----------
        topic : bytes
            define why message has been send
        msg: a Pickel dumps message
        """
        if topic == b'GUI_GRAPH':
            config_time, receiv_steps = msg
            if config_time != self.config_time:
                self.full_change(receiv_steps)
                self.config_time = config_time
        # Stager or Producer or Consumer state changes
        elif topic == b'GUI_ROUTER_CHANGE':
            self.router_change(msg)


    def full_change(self,receiv_steps):
        """
        Update self.steps with new receiv_steps
        Test if receiv_steps is same as self.steps.
        If yes do nothing otherwise set self.steps to receiv_steps

        Parameters
        -----------
        receiv_steps: list of StagerRep
        """
        if not self.steps:
            self.steps = receiv_steps
        else:
            nb_corresponding_step = 0
            for new_step in receiv_steps:
                for step in self.steps:

                    if step.name == new_step.name:
                        step.nb_job_done = new_step.nb_job_done
                        step.running = new_step.running
                        nb_corresponding_step+=1
                        break
            if  nb_corresponding_step != len(self.steps):
                # in case of pipeline configuration change and GUI
                # is not restarted
                self.steps = receiv_steps


    def step_change(self, msg):
        """Find which pipeline step has changed, and update its corresponding
        StagerRep
        Parameters
        ----------
        msg: list
            contains step name, step running flag and step nb_job_done
            receiv_steps: list of GUIStepInfo describing pipeline contents
        """
        name, running , nb_job_done = msg
        for step in self.steps:
            if step.name == name:
                step.running = running
                return

    def router_change(self, msg):
        """Called by GuiConnexion instance when it receives zmq message from pipeline
        Update pipeline state (self.steps) and force to update drawing
        Parameters
        ----------
        msg : a Pickle.dumps message
        """
        name, queue = msg
        name = name.split('_router')[0]
        step = self.get_step_by_name(name)
        if step:
            step.queue_length = queue



    def get_step_by_name(self, name):
        ''' Find a PipeStep in self.producer_step or  self.stager_steps or
        self.consumer_step
        Return: PipeStep if found, otherwise None
        '''
        for step in self.steps:
            if step.name == name or step.name.split('$$process')[0] == name:
                return step
        return None

    def finish(self):
        """ Force to stop this thread
        """
        self.stop = True

    def flow_has_finish(self):
        """ Flow finished. Set all step.runnng to False
        """
        for step in self.steps:
            step.running=0
        self.full_change(self.steps)
        self.message.emit(self.steps)

    def send_mode(self,msg):
        """ Flow can run in sequetial or multiprocess mode.
        This will informs InfoLabel of Flow MODE

        Parameters
        ----------
        msg: A Picle.dumps msg containing Flow mode
        """
        self.mode_message.emit(msg)



"""
dump a table of the event times and trigger patterns from a
simtelarray input file.
"""

import numpy as np
from astropy import units as u
from astropy.table import Table

import ctapipe.io.hessio as hessio
from ctapipe.core import Provenance, ToolConfigurationError
from ctapipe.core.traits import (Unicode, Dict, Bool)
from ..core import Tool

MAX_TELS = 1000


class DumpTriggersTool(Tool):
    description = Unicode(__doc__)
    name = 'ctapipe-dump-triggers'

    # =============================================
    # configuration parameters:
    # =============================================

    infile = Unicode(help='input simtelarray file').tag(config=True)

    outfile = Unicode('triggers.fits',
                      help='output filename (*.fits, *.h5)').tag(config=True)

    overwrite = Bool(False,
                     help="overwrite existing output file"
                     ).tag(config=True)

    # =============================================
    # map low-level options to high-level command-line options
    # =============================================

    aliases = Dict({'infile': 'DumpTriggersTool.infile',
                    'outfile': 'DumpTriggersTool.outfile'})

    flags = Dict({'overwrite': ({'DumpTriggersTool': {'overwrite': True}},
                                'Enable overwriting of output file')})

    examples = ('ctapipe-dump-triggers --infile gamma.simtel.gz '
                '--outfile trig.fits --overwrite'
                '\n\n'
                'If you want to see more output, use --log_level=DEBUG')



    # =============================================
    # The methods of the Tool (initialize, start, finish):
    # =============================================

    def add_event_to_table(self, event):
        """
        add the current hessio event to a row in the `self.events` table
        """
        gpstime = event.trig.gps_time

        if self._prev_gpstime is None:
            self._prev_gpstime = gpstime

        if self._current_starttime is None:
            self._current_starttime = gpstime

        relative_time = gpstime - self._current_starttime
        delta_t = gpstime - self._prev_gpstime
        self._prev_gpstime = gpstime

        # build the trigger pattern as a fixed-length array
        # (better for storage in FITS format)
        #trigtels = event.get_telescope_with_data_list()
        trigtels = event.dl0.tels_with_data
        self._current_trigpattern[:] = 0  # zero the trigger pattern
        self._current_trigpattern[list(trigtels)] = 1  # set the triggered tels
        # to 1

        # insert the row into the table
        self.events.add_row((event.dl0.event_id, relative_time.sec, delta_t.sec,
                             len(trigtels),
                             self._current_trigpattern))

    def setup(self):
        """ setup function, called before `start()` """

        if self.infile == '':
            raise ToolConfigurationError("No 'infile' parameter was specified. ")

        self.events = Table(names=['EVENT_ID', 'T_REL', 'DELTA_T',
                                   'N_TRIG', 'TRIGGERED_TELS'],
                            dtype=[np.int64, np.float64, np.float64,
                                   np.int32, np.uint8])

        self.events['TRIGGERED_TELS'].shape = (0, MAX_TELS)
        self.events['T_REL'].unit = u.s
        self.events['T_REL'].description = 'Time relative to first event'
        self.events['DELTA_T'].unit = u.s
        self.events.meta['INPUT'] = self.infile

        self._current_trigpattern = np.zeros(MAX_TELS)
        self._current_starttime = None
        self._prev_gpstime = None


    def start(self):
        """ main event loop """
        source = hessio.hessio_event_source(self.infile)

        for event in source:
            self.add_event_to_table(event)

    def finish(self):
        """
        finish up and write out results (called automatically after
        `start()`)
        """
        # write out the final table
        try:
            if self.outfile.endswith('fits') or self.outfile.endswith('fits.gz'):
                self.events.write(self.outfile, overwrite=self.overwrite)
            elif self.outfile.endswith('h5'):
                self.events.write(self.outfile, path='/events',
                                  overwrite=self.overwrite)
            else:
                self.events.write(self.outfile)

            Provenance().add_output_file(self.outfile)
        except IOError as err:
            self.log.warn("Couldn't write output (%s)", err)

        self.log.info('\n %s', self.events)


def main():
    tool = DumpTriggersTool()
    tool.run()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This packages contains python packages that are bundled with the affiliated
package but are external to the affiliated package, and hence are developed in
a separate source tree. Note that this package is distinct from the /cextern
directory of the source code distribution, as that directory only contains C
extension code.
"""



#from PyQt4.QtCore import *
from PyQt4.QtGui import QLabel

class InfoLabel(QLabel):
    """
    Displays steps name, queues size and numbers of job done
    """
    def __init__(self, *args):
        QLabel.__init__(self)
        self.data = dict()
        self.mode = 'sequential'

    def pipechange(self, steps):
        """Called by GuiConnexion instance when it receives zmq message from
        pipeline Update pipeline state (self.steps) and force to
        update drawing

        Parameters
        ----------
        topic : str

        """
        if steps:
            if self.mode == 'sequential':
                text ='  STEP                                       DONE    \n\n'
            else:
                text ='  STEP                                       RUNNING   QUEUE       DONE    \n\n'
            for step in steps:
                text+='{:-<80}\n'.format('-')
                text+=self.formatText(step)+'\n'
            self.setText(text)

    def formatText(self,step):
        """ Format Step state

        Parameters
        ----------
        step : StagerRep to format

        Returns
        -------
        A str containg sdtep state
        """
        text = str()
        step_name = step.name.split('$$process')[0]
        if len(step_name) < 18:
            name = ' ' + step_name + ((18-len(step_name))*2)*' '
        elif len(step_name) > 18:
            name =  ' ' +step_name[0:16] + '...'
        else:
            name =  ' ' + step_name
        queue = step.queue_length
        done = step.nb_job_done
        nb_proc = step.nb_process
        running = step.running
        if self.mode == 'sequential':
            text = name+ '\t' + str(done)
        else:
            text = name+ '\t' + str(running) + '\\' + str(nb_proc) + '\t' + str(queue) + '\t' + str(done)
        return text

    def reset(self):
        self.setText("")

    def mode_receive(self,mode):
        """
        Parameters
        ----------
        Flow mode (sequential or multiprocess)
        """
        self.mode = mode



from collections import defaultdict
from copy import deepcopy
from pprint import pformat
from textwrap import wrap


class Field:
    """
    Class for storing data in `Containers`.

    Parameters
    ----------
    default:
        default value of the item (this will be set when the `Container`
        is constructed, as well as when  `Container.reset()` is called
    description: str
        Help text associated with the item
    unit: `astropy.units.Quantity`
        unit to convert to when writing output, or None for no conversion
    ucd: str
        universal content descriptor (see Virtual Observatory standards)
    """

    def __init__(self, default, description="", unit=None, ucd=None):
        self.default = default
        self.description = description
        self.unit = unit

    def __repr__(self):
        desc = '{}'.format(self.description)
        if self.unit is not None:
            desc += ' [{}]'.format(self.unit)
        return desc


class ContainerMeta(type):
    '''
    The MetaClass for the Containers

    It reserves __slots__ for every class variable,
    that is of instance `Field` and sets all other class variables
    as read-only for the instances.

    This makes sure, that the metadata is immutable,
    and no new fields can be added to a container by accident.
    '''
    def __new__(cls, name, bases, dct):
        items = [
            k for k, v in dct.items()
            if isinstance(v, Field)
        ]
        dct['__slots__'] = tuple(items + ['meta'])
        dct['fields'] = {}

        for k in items:
            dct['fields'][k] = dct.pop(k)

        return type.__new__(cls, name, bases, dct)


class Container(metaclass=ContainerMeta):
    """Generic class that can hold and accumulate data to be passed
    between Components.

    The purpose of this class is to provide a flexible data structure
    that works a bit like a dict or blank Python class, but prevents
    the user from accessing members that have not been defined a
    priori (more like a C struct), and also keeps metdata information
    such as a description, defaults, and units for each item in the
    container.

    Containers can transform the data into a `dict` using the `
    Container.as_dict()` method.  This allows them to be written to an
    output table for example, where each Field defines a column. The
    `dict` conversion can be made recursively and even flattened so
    that a nested set of `Containers` can be translated into a set of
    columns in a flat table without naming conflicts (the name of the
    parent Field is pre-pended).

    Only members of instance `Field` will be used as output.
    For hierarchical data structures, Field can use `Container`
    subclasses or a `Map` as the default value.

    You should not make class hierarchies of Containers and only ever
    subclass the Container base class

    >>>    class MyContainer(Container):
    >>>        x = Field(100,"The X value")
    >>>        energy = Field(-1, "Energy measurement", unit=u.TeV)
    >>>
    >>>    cont = MyContainer()
    >>>    print(cont.x)
    100
    >>>    # metdata will become header keywords in an output file:
    >>>    cont.meta['KEY'] = value

    `Field`s inside `Containers` can contain instances of other
    `Containers`, to allow for a hierarchy of containers, and can also
    contain a `Map` for the case where one wants e.g. a set of
    sub-classes indexed by a value like the `telescope_id`. Examples
    of this can be found in `ctapipe.io.containers`

    `Containers` work by shadowing all class variables (which must be
    instances of `Field`) with instance variables of the same name the
    hold the value expected. If `Container.reset()` is called, all
    instance variables are reset to their default values as defined in
    the class.

    Finally, `Containers` can have associated metadata via their
    `meta` attribute, which is a `dict` of keywords to values.

    """
    def __init__(self, **fields):

        self.meta = {}
        for k, v in self.fields.items():
            setattr(self, k, deepcopy(v.default))

        for k, v in fields.items():
            setattr(self, k, v)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def items(self):
        """Generator over (key, value) pairs for the items"""
        return ((k, getattr(self, k)) for k in self.fields.keys())

    def as_dict(self, recursive=False, flatten=False):
        """
        convert the `Container` into a dictionary

        Parameters
        ----------
        recursive: bool
            sub-Containers should also be converted to dicts
        flatten: type
            return a flat dictionary, with any sub-field keys generated
            by appending the sub-Container name.
        """
        if not recursive:
            return dict(self.items())
        else:
            d = dict()
            for key, val in self.items():
                if key.startswith("_"):
                    continue
                if isinstance(val, Container) or isinstance(val, Map):
                    if flatten:
                        d.update({"{}_{}".format(key, k): v
                                  for k, v in val.as_dict(recursive).items()})
                    else:
                        d[key] = val.as_dict(recursive=recursive,
                                             flatten=flatten)
                    continue
                d[key] = val
            return d

    @classmethod
    def disable_attribute_check(cls):
        """
        Globally turn off attribute checking for all Containers,
        which provides a ~5-10x speed up for setting attributes.
        This may be used e.g. after code is tested to speed up operation.
        """
        cls.__setattr__ = object.__setattr__

    def reset(self, recursive=True):
        """ set all values back to their default values"""
        for name, value in self.fields.items():
            if isinstance(value, Container):
                if recursive:
                    getattr(self, name).reset()
            else:
                setattr(self, name, deepcopy(self.fields[name].default))

    def update(self, **values):
        """
        update more than one parameter at once (e.g. `update(x=3,y=4)`
        or `update(**dict_of_values)`)
        """
        for key in values:
            self[key] = values[key]

    def __str__(self):
        return pformat(self.as_dict(recursive=True))

    def __repr__(self):
        text = ["{}.{}:".format(type(self).__module__, type(self).__name__)]
        for name, item in self.fields.items():
            extra = ""
            if isinstance(getattr(self, name), Container):
                extra = ".*"
            if isinstance(getattr(self, name), Map):
                extra = "[*]"
            desc = "{:>30s}: {}".format(name + extra, repr(item))
            lines = wrap(desc, 80, subsequent_indent=' ' * 32)
            text.extend(lines)
        return "\n".join(text)


class Map(defaultdict):
    """A dictionary of sub-containers that can be added to a Container. This
    may be used e.g. to store a set of identical sub-Containers (e.g. indexed
    by `tel_id` or algorithm name).
    """

    def as_dict(self, recursive=False, flatten=False):
        if not recursive:
            return dict(self.items())
        else:
            d = dict()
            for key, val in self.items():
                if isinstance(val, Container) or isinstance(val, Map):
                    if flatten:
                        d.update({"{}_{}".format(key, k): v
                                  for k, v in val.as_dict(recursive).items()})
                    else:
                        d[key] = val.as_dict(recursive=recursive,
                                             flatten=flatten)
                    continue
                d[key] = val
            return d

    def reset(self, recursive=True):
        for key, val in self.items():
            if isinstance(val, Container):
                val.reset(recursive=True)



from ctapipe.instrument import TelescopeDescription
from astropy import units as u
import numpy as np

def test_telescope_description():

    # setup a dummy telescope that look like an MST with FlashCam
    foclen = 16*u.m
    pix_x = np.arange(1764, dtype=np.float) * u.m
    pix_y = np.arange(1764, dtype=np.float) * u.m

    tel = TelescopeDescription.guess(pix_x, pix_y, foclen)

    assert tel.camera.cam_id == 'FlashCam'
    assert tel.optics.tel_type == 'MST'
    assert str(tel) == 'MST:FlashCam'


import numpy as np
import logging

log = logging.getLogger(__name__)


def mean_squared_error(pixel_x, pixel_y, weights, radius, center_x, center_y):
    '''
    Calculate the weighted mean squared error for a circle

    Parameters
    ----------
    pixel_x: array-like
        x coordinates of the camera pixels
    pixel_y: array-like
        y coordinates of the camera pixels
    weights: array-like
        weights for the camera pixels, will usually be the pe charges
    radius: float
        radius of the ring
    center_x: float
        x coordinate of the ring center
    center_y: float
        y coordinate of the ring center
    '''
    r = np.sqrt((center_x - pixel_x)**2 + (center_y - pixel_y)**2)
    return np.average((r - radius)**2, weights=weights)


def photon_ratio_inside_ring(
        pixel_x, pixel_y, weights, radius, center_x, center_y, width
        ):
    '''
    Calculate the ratio of the photons inside a given ring with
    coordinates (center_x, center_y), radius and width.

    The ring is assumed to be in [radius - 0.5 * width, radius + 0.5 * width]

    Parameters
    ----------
    pixel_x: array-like
        x coordinates of the camera pixels
    pixel_y: array-like
        y coordinates of the camera pixels
    weights: array-like
        weights for the camera pixels, will usually be the pe charges
    radius: float
        radius of the ring
    center_x: float
        x coordinate of the ring center
    center_y: float
        y coordinate of the ring center
    width: float
        width of the ring
    '''

    total = np.sum(weights)

    pixel_r = np.sqrt((center_x - pixel_x)**2 + (center_y - pixel_y)**2)
    mask = np.logical_and(
        pixel_r >= radius - 0.5 * width,
        pixel_r <= radius + 0.5 * width
    )

    inside = np.sum(weights[mask])

    return inside / total


def ring_completeness(
        pixel_x,
        pixel_y,
        weights,
        radius,
        center_x,
        center_y,
        threshold=30,
        bins=30,
        ):
    '''
    Estimate how complete a ring is.
    Bin the light distribution along the the ring and apply a threshold to the
    bin content.

    Parameters
    ----------
    pixel_x: array-like
        x coordinates of the camera pixels
    pixel_y: array-like
        y coordinates of the camera pixels
    weights: array-like
        weights for the camera pixels, will usually be the pe charges
    radius: float
        radius of the ring
    center_x: float
        x coordinate of the ring center
    center_y: float
        y coordinate of the ring center
    threshold: float
        number of photons a bin must contain to be counted
    bins: int
        number of bins to use for the histogram

    Returns
    -------
    ring_completeness: float
        the ratio of bins above threshold

    Returns
    -------
    ring_completeness: float
        the ratio of bins above threshold
    '''

    angle = np.arctan2(pixel_y - center_y, pixel_x - center_x)

    hist, edges = np.histogram(angle, bins=bins, range=[-np.pi, np.pi], weights=weights)

    bins_above_threshold = hist > threshold

    return np.sum(bins_above_threshold) / bins



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Coordinates.
"""


from astropy.utils import iers
iers.conf.auto_download = True # auto-fetch updates to IERS_A table


from .angular_frames import *
from .ground_frames import *



"""
Pedestal Calculation Functions
"""

import numpy as np


def calc_pedestals_from_traces(traces, start_sample, end_sample):
    """A very simple algorithm to calculates pedestals and pedestal
    variances from camera traces by integrating the samples over a
    fixed window for all pixels.  This assumes that the data are
    sample-mode (e.g. cameras that return time traces for each pixel).

    Parameters
    ----------

    traces: array of shape (n_pixels, n_samples)
        time-sampled camera data in a 2D array pixel x sample
    start_sample: int
        index of starting sample over which to integrate
    end_sample: int
        index of ending sample over which to integrate

    Returns
    -------

    two arrays of length n_pix (the first dimension of the input trace
    array). The first array contains the pedestal values, and the
    second is the pedestal variances over the sample window.

    """
    traces = np.asanyarray(traces)  # ensure this is an ndarray
    peds = traces[:, start_sample:end_sample].mean(axis=1)
    pedvars = traces[:, start_sample:end_sample].var(axis=1)
    return peds, pedvars


# PUT OTHER PEDESTAL CALCULATION FUNCTIONS HERE:
# ----------------------------------------------



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import pytest
from ctapipe.utils.fitshistogram import Histogram

def compare_histograms(hist1: Histogram, hist2: Histogram):
    """ check that 2 histograms are identical in value """
    assert hist1.ndims == hist2.ndims
    assert (hist1.axis_names == hist2.axis_names).all()
    assert (hist1.data == hist2.data).all

    for ii in range(hist1.ndims):
        assert np.isclose(hist1.bin_lower_edges[ii],
                          hist2.bin_lower_edges[ii]).all()


def test_histogram_str():
    hist = Histogram(nbins=[5, 10],
                     ranges=[[-2.5, 2.5], [-1, 1]], name="testhisto")
    expected = ("Histogram(name='testhisto', axes=['axis0', 'axis1'], "
                "nbins=[ 5 10], ranges=[[-2.5  2.5]\n [-1.   1. ]])")
    assert str(hist) == expected


def test_histogram_fill_and_read():

    hist = Histogram(nbins=[5, 10], ranges=[[-2.5, 2.5], [-1, 1]])

    pa = (0.1, 0.1)
    pb = (-0.55, 0.55)

    a = np.ones((100, 2)) * pa  # point at 0.1,0.1
    b = np.ones((10, 2)) * pb  # 10 points at -0.5,0.5

    hist.fill(a)
    hist.fill(b)

    va = hist.get_value(pa)[0]
    vb = hist.get_value(pb)[0]

    assert va == 100
    assert vb == 10


def test_histogram_range_fill_and_read():
    """
    Check that the correct bin is read and written for multiple
    binnings and fill positions
    """

    num = 100

    for nxbins in np.arange(1, 50, 1):
        for xx in np.arange(-2.0, 2.0, 0.1):
            pp = (xx + 0.01829384, 0.1)
            coords = np.ones((num, 2)) * np.array(pp)
            hist = Histogram(nbins=[nxbins, 10],
                             ranges=[[-2.5, 2.5], [-1, 1]])
            hist.fill(coords)
            val = hist.get_value(pp)[0]
            assert val == num
            del hist


def test_outliers():
    """
    Check that out-of-range values work as expected
    """
    H = Histogram(nbins=[5, 10], ranges=[[-2.5, 2.5], [-1, 1]])
    H.fill(np.array([[1, 1], ]))
    val1 = H.get_value((100, 100), outlier_value=-10000)
    val2 = H.get_value((-100, 0), outlier_value=None)
    assert val1 == -10000
    assert val2 == 0


@pytest.fixture(scope='session')
def histogram_file(tmpdir_factory):
    """ a fixture that fetches a temporary output dir/file for a test
    histogram"""
    return str(tmpdir_factory.mktemp('data').join('histogram_test.fits'))


def test_histogram_fits(histogram_file):
    """
    Write to fits,read back, and check
    """

    hist = Histogram(nbins=[5, 11], ranges=[[-2.5, 2.5], [-1, 1]])
    hist.fill(np.array([[0, 0],
                        [0, 0.5]]))

    hist.to_fits().writeto(histogram_file, clobber=True)
    newhist = Histogram.from_fits(histogram_file)

    # check that the values are the same
    compare_histograms(hist, newhist)


def test_histogram_resample_inplace():
    hist = Histogram(nbins=[5, 11], ranges=[[-2.5, 2.5], [-1, 1]])
    hist.fill(np.array([[0, 0],
                     [0,0.5]]))

    for testpoint in [(0,0), (0,1), (1,0), (3,3)]:
        val0 = hist.get_value(testpoint)
        hist.resample_inplace((10, 22))
        val1 = hist.get_value(testpoint)
        hist.resample_inplace((5, 11))
        val2 = hist.get_value(testpoint)

        # at least check the resampling is undoable
        assert np.isclose(val0[0], val2[0])


"""This module defines the important coordinate systems to be used in
reconstruction with the CTA pipeline and the transformations between
this different systems. Frames and transformations are defined using
the astropy.coordinates framework. This module contains transformations
between angular systems.

For examples on usage see examples/coordinate_transformations.py

This code is based on the coordinate transformations performed in the
read_hess code

TODO:

- Tests Tests Tests!
- Check cartesian system is still accurate for the nominal and
  telescope systems (may need a spherical system)
- Benchmark transformation time
- should use `astropy.coordinates.Angle` for all angles here
"""

import astropy.units as u
import numpy as np
from astropy.coordinates import (BaseCoordinateFrame,
                                 CartesianRepresentation,
                                 UnitSphericalRepresentation,
                                 FunctionTransform, RepresentationMapping)

try:
    # FrameAttribute was renamed Attribute in astropy 2.0
    # TODO: should really use subclasses like QuantityAttribute
    from astropy.coordinates import FrameAttribute as Attribute
except ImportError:
    from astropy.coordinates import Attribute

from astropy.coordinates import frame_transform_graph
from numpy import cos, sin, arctan, arctan2, arcsin, sqrt, arccos, tan
from ..coordinates.representation import PlanarRepresentation

__all__ = [
    'CameraFrame',
    'TelescopeFrame',
    'NominalFrame',
    'HorizonFrame'
]


class CameraFrame(BaseCoordinateFrame):
    """Camera coordinate frame.  The camera frame is a simple physical
    cartesian frame, describing the 2 dimensional position of objects
    in the focal plane of the telescope Most Typically this will be
    used to describe the positions of the pixels in the focal plane

    Frame attributes:

    * ``focal_length``
        Focal length of the telescope as a unit quantity (usually meters)
    * ``rotation``
        Rotation angle of the camera (0 deg in most cases)
    """
    default_representation = CartesianRepresentation
    focal_length = Attribute(default=None)
    rotation = Attribute(default=0 * u.deg)
    pointing_direction = Attribute(default=None)
    array_direction = Attribute(default=None)


class TelescopeFrame(BaseCoordinateFrame):
    """Telescope coordinate frame.  Cartesian system to describe the
    angular offset of a given position in reference to pointing
    direction of a given telescope When pointing corrections become
    available they should be applied to the transformation between
    this frame and the camera frame

    Frame attributes:

    * ``focal_length``
        Focal length of the telescope as a unit quantity (usually meters)
    * ``rotation``
        Rotation angle of the camera (0 deg in most cases)
    * ``pointing_direction``
        Alt,Az direction of the telescope pointing

    """
    default_representation = PlanarRepresentation
    pointing_direction = Attribute(default=None)


class NominalFrame(BaseCoordinateFrame):
    """Nominal coordinate frame.  Cartesian system to describe the angular
    offset of a given position in reference to pointing direction of a
    nominal array pointing position. In most cases this frame is the
    same as the telescope frame, however in the case of divergent
    pointing they will differ.  Event reconstruction should be
    performed in this system

    Frame attributes:
    
    * ``array_direction``
        Alt,Az direction of the array pointing
    * ``pointing_direction``
        Alt,Az direction of the telescope pointing

    """
    default_representation = PlanarRepresentation
    pointing_direction = Attribute(default=None)
    array_direction = Attribute(default=None)


class HorizonFrame(BaseCoordinateFrame):
    """Horizon coordinate frame. Spherical system used to describe the direction
    of a given position, in terms of the altitude and azimuth of the system. In
    practice this is functionally identical as the astropy AltAz system, but this
    implementation allows us to pass array pointing information, allowing us to directly
    transform to the Horizon Frame from the Camera system.
    The Following attributes are carried over from the telescope frame
    to allow a direct transformation from the camera frame

    Frame attributes:

    * ``array_direction``
        Alt,Az direction of the array pointing
    * ``pointing_direction``
        Alt,Az direction of the telescope pointing

    """
    default_representation = UnitSphericalRepresentation

    frame_specific_representation_info = {
        'spherical': [RepresentationMapping('lon', 'az'),
                      RepresentationMapping('lat', 'alt')],
    }

    frame_specific_representation_info['unitspherical'] = frame_specific_representation_info['spherical']

    pointing_direction = Attribute(default=None)
    array_direction = Attribute(default=None)


# Transformations defined below this point

def altaz_to_offset(obj_azimuth, obj_altitude, azimuth, altitude):
    """
    Function to convert a given altitude and azimuth to a cartesian angular
    angular offset with regard to a give reference system
    (This function is directly lifted from read_hess)

    Parameters
    ----------
    obj_azimuth: float
        Event azimuth (radians)
    obj_altitude: float
        Event altitude (radians)
    azimuth: float
        Reference system azimuth (radians)
    altitude: float
        Reference system altitude (radians)

    Returns
    -------
    x_off,y_off: (float,float)
        Offset of the event in the reference system (in radians)
    """

    diff_az = obj_azimuth - azimuth
    cosine_obj_alt = cos(obj_altitude)

    xp0 = -cos(diff_az) * cosine_obj_alt
    yp0 = sin(diff_az) * cosine_obj_alt
    zp0 = sin(obj_altitude)

    sin_sys_alt = sin(altitude)
    cos_sys_alt = cos(altitude)

    xp1 = sin_sys_alt * xp0 + cos_sys_alt * zp0
    yp1 = yp0
    zp1 = -cos_sys_alt * xp0 + sin_sys_alt * zp0

    disp = tan(arccos(zp1))
    alpha = arctan2(yp1, xp1)

    x_off = disp * cos(alpha)
    y_off = disp * sin(alpha)

    return x_off, y_off


def offset_to_altaz(x_off, y_off, azimuth, altitude):
    """Function to convert an angular offset with regard to a give
    reference system to an an absolute altitude and azimuth (This
    function is directly lifted from read_hess)

    Parameters
    ----------
    x_off: float
        X offset of the event in the reference system
    y_off: float
        Y offset of the event in the reference system
    azimuth: float
        Reference system azimuth (radians)
    altitude: float
        Reference system altitude (radians)

    Returns
    -------
    obj_altitude,obj_azimuth: (float,float)
        Absolute altitude and azimuth of the event
    """

    unit = azimuth.unit

    x_off = x_off.to(u.rad).value
    y_off = y_off.to(u.rad).value
    azimuth = azimuth.to(u.rad).value
    altitude = altitude.to(u.rad).value

    offset = sqrt(x_off * x_off + y_off * y_off)
    pos = np.where(offset == 0)  # find offset 0 positions
    if len(pos[0]) > 0:
        offset[pos] = 1e-12  # add a very small offset to prevent math errors

    atan_off = arctan(offset)

    sin_atan_off = sin(atan_off)
    xp1 = x_off * (sin_atan_off / offset)
    yp1 = y_off * (sin_atan_off / offset)
    zp1 = cos(atan_off)

    sin_obj_alt = sin(altitude)
    cos_obj_alt = cos(altitude)

    xp0 = sin_obj_alt * xp1 - cos_obj_alt * zp1
    yp0 = yp1
    zp0 = cos_obj_alt * xp1 + sin_obj_alt * zp1

    obj_altitude = arcsin(zp0)
    obj_azimuth = arctan2(yp0, -xp0) + azimuth

    if len(pos[0]) > 0:
        obj_altitude[pos] = altitude
        obj_azimuth[pos] = azimuth

    obj_altitude = obj_altitude * u.rad
    obj_azimuth = obj_azimuth * u.rad

    # if obj_azimuth.value < 0.:
    #    obj_azimuth += 2.*pi
    # elif obj_azimuth.value >= (2.*pi ):
    #    obj_azimuth -= 2.*pi

    return obj_altitude.to(unit), obj_azimuth.to(unit)


# Transformation between nominal and AltAz system


@frame_transform_graph.transform(FunctionTransform, NominalFrame, HorizonFrame)
def nominal_to_altaz(norm_coord, altaz_coord):
    """
    Transformation from nominal system to astropy AltAz system

    Parameters
    ----------
    norm_coord: `astropy.coordinates.SkyCoord`
        nominal system
    altaz_coord: `astropy.coordinates.SkyCoord`
        AltAz system

    Returns
    -------
    AltAz Coordinates
    """
    alt_norm, az_norm = norm_coord.array_direction.alt, norm_coord.array_direction.az

    if type(norm_coord.x.value).__module__ != np.__name__:
        x_off = np.zeros(1)
        x_off[0] = norm_coord.x.value
        x_off = x_off * norm_coord.x.unit
        y_off = np.zeros(1)
        y_off[0] = norm_coord.y.value
        y_off = y_off * norm_coord.y.unit
    else:
        x_off = norm_coord.x
        y_off = norm_coord.y

    altitude, azimuth = offset_to_altaz(x_off, y_off, az_norm, alt_norm)

    representation = UnitSphericalRepresentation(lon=azimuth.to(u.deg), lat=altitude.to(u.deg))
    return altaz_coord.realize_frame(representation)


@frame_transform_graph.transform(FunctionTransform, HorizonFrame, NominalFrame)
def altaz_to_nominal(altaz_coord, norm_coord):
    """
    Transformation from astropy AltAz system to nominal system

    Parameters
    ----------
    altaz_coord: `astropy.coordinates.SkyCoord`
        AltAz system
    norm_coord: `astropy.coordinates.SkyCoord`
        nominal system

    Returns
    -------
    nominal Coordinates
    """
    alt_norm, az_norm = norm_coord.array_direction.alt, norm_coord.array_direction.az
    azimuth = altaz_coord.az
    altitude = altaz_coord.alt
    x_off, y_off = altaz_to_offset(azimuth, altitude, az_norm, alt_norm)
    x_off = x_off * u.rad
    y_off = y_off * u.rad
    representation = PlanarRepresentation(
        x_off.to(u.deg), y_off.to(u.deg))

    return norm_coord.realize_frame(representation)


# Transformation between telescope and nominal frames


@frame_transform_graph.transform(FunctionTransform, TelescopeFrame, NominalFrame)
def telescope_to_nominal(tel_coord, norm_frame):
    """
    Coordinate transformation from telescope frame to nominal frame

    Parameters
    ----------
    tel_coord: `astropy.coordinates.SkyCoord`
        TelescopeFrame system
    norm_frame: `astropy.coordinates.SkyCoord`
        NominalFrame system

    Returns
    -------
    NominalFrame coordinates
    """
    alt_tel, az_tel = tel_coord.pointing_direction.alt, tel_coord.pointing_direction.az
    alt_norm, az_norm = norm_frame.array_direction.alt, norm_frame.array_direction.az
    alt_trans, az_trans = offset_to_altaz(
        tel_coord.x, tel_coord.y, az_tel, alt_tel)

    x_off, y_off = altaz_to_offset(az_trans, alt_trans, az_norm, alt_norm)
    x_off = x_off * u.rad
    y_off = y_off * u.rad

    representation = PlanarRepresentation(
        x_off.to(tel_coord.x.unit), y_off.to(tel_coord.x.unit))

    return norm_frame.realize_frame(representation)


@frame_transform_graph.transform(FunctionTransform, NominalFrame, TelescopeFrame)
def nominal_to_telescope(norm_coord, tel_frame):
    """
    Coordinate transformation from nominal to telescope system

    Parameters
    ----------
    norm_coord: `astropy.coordinates.SkyCoord`
        NominalFrame system
    tel_frame: `astropy.coordinates.SkyCoord`
        TelescopeFrame system

    Returns
    -------
    TelescopeFrame coordinates

    """
    alt_tel, az_tel = tel_frame.pointing_direction.alt, tel_frame.pointing_direction.az
    alt_norm, az_norm = norm_coord.array_direction.alt, norm_coord.array_direction.az

    alt_trans, az_trans = offset_to_altaz(
        norm_coord.x, norm_coord.y, az_norm, alt_norm)
    x_off, y_off = altaz_to_offset(az_trans, alt_trans, az_tel, alt_tel)
    x_off = x_off * u.rad
    y_off = y_off * u.rad

    representation = PlanarRepresentation(x_off.to(norm_coord.x.unit), y_off.to(norm_coord.x.unit))

    return tel_frame.realize_frame(representation)


# Transformations between camera frame and telescope frame
@frame_transform_graph.transform(FunctionTransform, CameraFrame, TelescopeFrame)
def camera_to_telescope(camera_coord, telescope_frame):
    """
    Transformation between CameraFrame and TelescopeFrame

    Parameters
    ----------
    camera_coord: `astropy.coordinates.SkyCoord`
        CameraFrame system
    telescope_frame: `astropy.coordinates.SkyCoord`
        TelescopeFrame system
    Returns
    -------
    TelescopeFrame coordinate
    """
    x_pos = camera_coord.cartesian.x
    y_pos = camera_coord.cartesian.y

    rot = camera_coord.rotation
    if rot == 0:
        x_rotated = x_pos
        y_rotated = y_pos
    else:
        x_rotated = x_pos * cos(rot) - y_pos * sin(rot)
        y_rotated = x_pos * sin(rot) + y_pos * cos(rot)

    focal_length = camera_coord.focal_length

    x_rotated = (x_rotated / focal_length) * u.rad
    y_rotated = (y_rotated / focal_length) * u.rad
    representation = PlanarRepresentation(x_rotated, y_rotated)

    return telescope_frame.realize_frame(representation)


@frame_transform_graph.transform(FunctionTransform, TelescopeFrame, CameraFrame)
def telescope_to_camera(telescope_coord, camera_frame):
    """
    Transformation between TelescopeFrame and CameraFrame

    Parameters
    ----------
    telescope_coord: `astropy.coordinates.SkyCoord`
        TelescopeFrame system
    camera_frame: `astropy.coordinates.SkyCoord`
        CameraFrame system

    Returns
    -------
    CameraFrame Coordinates
    """
    x_pos = telescope_coord.cartesian.x
    y_pos = telescope_coord.cartesian.y
    # reverse the rotation applied to get to this system
    rot = camera_frame.rotation * -1

    if rot == 0:  # if no rotation applied save a few cycles
        x_rotated = x_pos
        y_rotated = y_pos
    else:  # or else rotate all positions around the camera centre
        x_rotated = x_pos * cos(rot) - y_pos * sin(rot)
        y_rotated = x_pos * sin(rot) + y_pos * cos(rot)

    focal_length = camera_frame.focal_length
    # Remove distance units here as we are using small angle approx
    x_rotated = x_rotated.to(u.rad) * (focal_length / u.m)
    y_rotated = y_rotated.to(u.rad) * (focal_length / u.m)

    representation = CartesianRepresentation(
        x_rotated.value * u.m, y_rotated.value * u.m, 0 * u.m)

    return camera_frame.realize_frame(representation)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from ctapipe.core import Container, Field, Map


def test_container():

    class ExampleContainer(Container):
        x = Field(-1, "x value")
        y = Field(-1, "y value")

    cont = ExampleContainer()
    cont2 = ExampleContainer()

    assert cont.x == -1
    assert cont.y == -1

    # test setting a value
    cont.x = 10
    assert cont.x == 10

    # make sure the value is set in the instance not the class
    assert cont2.x == -1

    # make sure you can't set an attribute that isn't defined
    with pytest.raises(AttributeError):
        cont.z = 100

    # test resetting values to defaults:
    cont.reset()
    assert cont.x == -1

    # test adding metadata
    cont.meta['stuff'] = 'things'
    assert 'stuff' in cont.meta and cont.meta['stuff'] == 'things'


def test_child_containers():
    class ChildContainer(Container):
        z = Field(1, "sub-item")

    class ParentContainer(Container):
        x = Field(0, "some value")
        child = Field(ChildContainer(), "a child")

    cont = ParentContainer()
    assert cont.child.z == 1


def test_map_containers():

    class ChildContainer(Container):
        z = Field(1, "sub-item")

    class ParentContainer(Container):
        x = Field(0, "some value")
        children = Field(Map(), "map of tel_id to child")

    cont = ParentContainer()
    cont.children[10] = ChildContainer()
    cont.children[5] = ChildContainer()

    cont.children[5].z = 99
    assert cont.children[5].z == 99

    cont.reset()
    assert 5 not in cont.children


def test_container_as_dict():

    class ChildContainer(Container):
        z = Field(1, "sub-item")

    class ParentContainer(Container):
        x = Field(0, "some value")
        child = Field(ChildContainer(), "a child")

    cont = ParentContainer()

    the_flat_dict = cont.as_dict(recursive=True, flatten=True)
    the_dict = cont.as_dict(recursive=True, flatten=False)

    assert 'child_z' in the_flat_dict
    assert 'child' in the_dict and 'z' in the_dict['child']


def test_container_brackets():

    class TestContainer(Container):
        answer = Field(-1, "The answer to all questions")

    t = TestContainer()

    t['answer'] = 42

    with pytest.raises(AttributeError):
        t['foo'] = 5



from multiprocessing import Process
from multiprocessing import Value
from zmq import POLLIN
from zmq import PUB
from zmq import ROUTER
from zmq import Poller
from zmq import Context
from zmq.error import ZMQError
from pickle import dumps
from pickle import loads
from ctapipe.core import Component

class RouterQueue(Process, Component):

    """`RouterQueue` class represents a router between pipeline steps, and it
    manages queue for prev step.
    It is derived from Process class.
    RouterQueue class is connected to one or more input steps and to one
    or more output steps thanks to zmq.ROUTER sockets.
    If inputs arrive quickers than output are sent (because next stage have not
    enough time to compute) these inputs are queued in RouterQueue.
    RouterQueue send output the next steps in LRU(last recently used) pattern.
    If a queue limit is reach for a step, RouterQueue replies "FULL"
    without taking new job in account, so the prev step have to send it again
    later.
    """
    def __init__(
        self, connections=None, gui_address=None):
        """
        Parameters
        ----------
        connections: dict {'STEP_NANE' : (STEP port in, STEP port out,
                                            STEP queue lengh max )}
            Port in, port out  for socket for each next steps.
            Max queue lengh (-1 menans no maximum)
        gui_address : str
            GUI port for ZMQ 'hostname': + 'port'
        """
        Process.__init__(self)
        Component.__init__(self,parent=None)
        self.gui_address = gui_address
        # list of available stages which receive next job
        # This list allow to use next_stage in a LRU (Last recently used)
        # pattern
        self.next_available_stages = dict()
        # queue jobs to be send to next_stage
        self.queue_jobs = dict()
        self.router_sockets = dict()
        self.dealer_sockets = dict()
        self.queue_limit = dict()
        self.connections = connections or {}
        self.done = False
        self._stop = Value('i',0)
        self._total_queue_size = Value('i',0)

    def run(self):
        """
        Method representing the process’s activity.
        It sends a job present in its queue (FIFO) to an available stager
        (if exist). Then it polls its sockets (in and out).
        When received new input from input socket, it appends contains to
        its queue.
        When received a signal from its output socket, it append sender
        (a pipeline step) to availble stagers list
        """
        if self.init_connections():
            nb_job_remains = 0
            while not self.stop or nb_job_remains > 0:
                for name in self.connections:
                    queue = self.queue_jobs[name]
                    next_available = self.next_available_stages[name]
                    if queue and next_available:
                        # get that oldest job and remove it form list
                        job = self.queue_jobs[name].pop(0)
                        if self.gui_address : self.update_gui(name)
                        # Get the next_stage for new job, and remove it from
                        # available list
                        next_stage = self.next_available_stages[name].pop(0)
                        # send new job
                        self.dealer_sockets[name].send_multipart(
                            [next_stage, b"", dumps(job)])
                    # check if new socket message arrive. Or skip after timeout
                    # (100 s)
                sockets = dict(self.poller.poll(100))
                # Test if message arrived from next_stages
                for n, socket_dealer in self.dealer_sockets.items():
                    if (socket_dealer in sockets and
                            sockets[socket_dealer] == POLLIN):
                        request = socket_dealer.recv_multipart()
                        # Get next_stage identity(to responde) and message
                        next_stage, _, _ = request[:3]
                        # add stager to next_available_stages
                        self.next_available_stages[n].append(next_stage)
                # Test if message arrived from prev_stage (stage or producer)
                for n, socket_router in self.router_sockets.items():
                    if (socket_router in sockets and
                            sockets[socket_router] == POLLIN):
                        # Get next prev_stage request
                        address, empty, request = socket_router.recv_multipart()
                        # store it to job queue
                        queue = self.queue_jobs[n]
                        if (len(queue) > self.queue_limit[n]
                        and self.queue_limit[n] != -1) :
                            socket_router.send_multipart([address, b"", b"FULL"])
                        else:
                            queue.append(loads(request))
                            if self.gui_address : self.update_gui(n)
                            socket_router.send_multipart([address, b"", b"OK"])
                nb_job_remains = 0
                for n, queue in self.queue_jobs.items():
                    nb_job_remains += len(queue)
                self._total_queue_size.value = nb_job_remains
            for socket in self.router_sockets.values():
                socket.close()
            for socket in self.dealer_sockets.values():
                socket.close()
        self.done = True

    def isQueueEmpty(self, stage_name=None):
        """
        Get status of steps' queue
        Parameters
        ----------
        stage_name: str
            router_name corresponding to stager name or consumer name
            If stage_name=None, it take in account all queues
        Returns
        -------
        True is corresponding queue is empty, otherwise False
        If stage_name = None it returns True if sum of all queues is 0
        """
        if stage_name:
            val = stage_name.find("$$process_number$$")
            if val != -1:
                name_to_search = stage_name[0:val]
            else:
                name_to_search = stage_name
            for name, queue in self.queue_jobs.items():
                if name.find("_router"):
                    pos = name.find("_router")
                    name = name[:pos]
                if name == name_to_search:
                    if not queue:
                        return True
            return False
        else:
            for queue in self.queue_jobs.values():
                if queue:
                    return False
            return True

    def init_connections(self):
        """
        Initialise zmq sockets, poller and queues.
        Because this class is s Process, This method must be call in the run
         method to be hold by the correct process.
        """
        # Prepare our context and sockets
        context = Context()
        # Socket to talk to prev_stages
        for name,connections in self.connections.items():
            self.queue_limit[name] = connections[2]
            sock_router = context.socket(ROUTER)
            try:
                sock_router.bind('tcp://*:' + connections[0])
            except ZMQError as e:
                self.log.error('{} : tcp://localhost:{}'
                               .format(e,  connections[0]))
                return False
            self.router_sockets[name] = sock_router
            # Socket to talk to next_stages
            sock_dealer = context.socket(ROUTER)
            try:
                sock_dealer.bind("tcp://*:" + connections[1] )
            except ZMQError as e:
                self.log.error('{} : tcp://localhost:{}'
                               .format(e,  connections[1]))
                return False
            self.dealer_sockets[name] = sock_dealer
            self.next_available_stages[name] = list()
            self.queue_jobs[name] = list()
        # Use a ZMQ Pool to get multichannel message
        self.poller = Poller()
        # Register dealer socket to next_stage
        for n, dealer in self.dealer_sockets.items():
            self.poller.register(dealer, POLLIN)
        for n, router in self.router_sockets.items():
            self.poller.register(router, POLLIN)
        # Register router socket to prev_stages or producer
        self.socket_pub = context.socket(PUB)
        if self.gui_address is not None:
            try:
                self.socket_pub.connect("tcp://" + self.gui_address)
            except ZMQError as e:
                self.log.error("".format(e, self.gui_address))
                return False
        # This flag stop this current process
        return True

    def update_gui(self, name):
        """
        send status to GUI
        name: str
            step name
        """
        msg = [name, str(len(self.queue_jobs[name]))]
        self.socket_pub.send_multipart(
            [b'GUI_ROUTER_CHANGE', dumps(msg)])

    @property
    def stop(self):
        return self._stop.value

    @stop.setter
    def stop(self, value):
        self._stop.value = value

    @property
    def total_queue_size(self):
        return self._total_queue_size.value



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# coding: utf8
from types import GeneratorType
from multiprocessing import Process
from multiprocessing import  Value
from pickle import loads
from zmq import POLLIN
from zmq import REQ
from zmq import Poller
from zmq import Context
from ctapipe.flow.multiprocess.connections import Connections
from ctapipe.core import Component

class StagerZmq(Component, Process, Connections):

    """`StagerZmq` class represents a Stager pipeline Step.
    It is derived from Process class.
    It receives new input from its prev stage, thanks to its ZMQ REQ socket,
    and executes its coroutine objet's run method by passing
    input as parameter. Finaly it sends coroutine returned value to its next
    stage, thanks to its ZMQ REQ socket,
    The process is launched by calling run method.
    init() method is call by run method.
    The process is stopped by setting share data stop to True
    """
    def __init__(
            self, coroutine, sock_job_for_me_port,
            name=None, connections=None, main_connection_name=None):
        """
        Parameters
        ----------
        coroutine : Class instance that contains init, run and finish methods
        sock_job_for_me_port: str
            Port number for input socket url
        name: str
            Stage name
        main_connection_name : str
            Default next step name. Used to send data when destination is not provided
        connections: dict {'STEP_NANE' : (zmq STEP_NANE port in)}
            Port number for socket for each next steps
        """
        Process.__init__(self)
        Component.__init__(self,parent=None)
        self.name = name
        Connections.__init__(self, main_connection_name, connections)
        self.coroutine = coroutine
        self.sock_job_for_me_url = 'tcp://localhost:' + sock_job_for_me_port
        self.done = False
        self.waiting_since = Value('i',0)
        self._nb_job_done = Value('i',0)
        self._stop = Value('i',0)
        self._running = Value('i',0)

    def init(self):
        """
        Initialise coroutine sockets and poller
        Returns
        -------
        True if coroutine init and init_connections methods returns True,
         otherwise False
        """
        if self.name is None:
            self.name = "STAGER"
        if self.coroutine is None:
            return False
        if self.coroutine.init() == False:
            return False

        self.coroutine.connections = list(self.connections)

        return self.init_connections()

    def run(self):
        """
        Method representing the process's activity.
        It polls its socket and when received new input from it,
        it executes coroutine run method by passing new input.
        Then it sends coroutine return value to its next stage,
        thanks to its ZMQ REQ socket.
        The poll method's timeout is 100 ms in case of self.stop flag
        has been set to False.
        Atfer the main while loop, coroutine.finish method is called
        """
        if self.init()  :
            while not self.stop:
                sockets = dict(self.poll.poll(100))  # Poll or time out (100ms)
                if (self.sock_for_me in sockets and
                        sockets[self.sock_for_me] == POLLIN):
                    #  Get the input from prev_stage
                    self.waiting_since.value = 0
                    self.running = 1
                    request = self.sock_for_me.recv_multipart()
                    receiv_input = loads(request[0])
                    # do the job
                    results = self.coroutine.run(receiv_input)
                    if isinstance(results, GeneratorType):
                        for val in results:
                            msg,destination = self.get_destination_msg_from_result(val)
                            self.send_msg(msg,destination)
                    else:
                        msg,destination = self.get_destination_msg_from_result(results)
                        self.send_msg(msg,destination)
                    # send acknoledgement to prev router/queue to inform it that I
                    # am available
                    self.sock_for_me.send_multipart(request)
                    self._nb_job_done.value = self._nb_job_done.value + 1
                    self.running = 0
                else:
                    self.waiting_since.value = self.waiting_since.value+100 # 100 ms
            self.sock_for_me.close()
        self.finish()
        self.done = True

    def finish(self):
        self.coroutine.finish()

    def init_connections(self):
        """
        Initialise zmq sockets.
        Because this class is s Process, This method must be call in the run
         method to be hold by the correct process.
        """
        Connections.init_connections(self)
        context = Context()
        self.sock_for_me = context.socket(REQ)
        self.sock_for_me.connect(self.sock_job_for_me_url)
        # Use a ZMQ Pool to get multichannel message
        self.poll = Poller()
        # Register sockets
        self.poll.register(self.sock_for_me, POLLIN)
        # Send READY to next_router to inform about my capacity to compute new
        # job
        self.sock_for_me.send_pyobj("READY")
        return True

    @property
    def wait_since(self):
        return self.waiting_since.value

    @property
    def stop(self):
        return self._stop.value

    @stop.setter
    def stop(self, value):
        self._stop.value = value

    @property
    def nb_job_done(self):
        return self._nb_job_done.value

    @nb_job_done.setter
    def nb_job_done(self, value):
        self._nb_job_done.value = value


    @property
    def running(self):
        return self._running.value

    @running.setter
    def running(self, value):
        self._running.value = value



from ctapipe.calib.camera import CameraCalibrator
from ctapipe.io.hessio import hessio_event_source
from ctapipe.utils import get_dataset
from numpy.testing import assert_allclose


def get_test_event():
    filename = get_dataset('gamma_test.simtel.gz')
    source = hessio_event_source(filename, requested_event=409,
                                 use_event_id=True)
    event = next(source)
    return event


def test_camera_calibrator():
    event = get_test_event()
    telid = 11

    calibrator = CameraCalibrator(None, None)

    calibrator.calibrate(event)
    image = event.dl1.tel[telid].image
    assert_allclose(image[0, 0], -2.216, 1e-3)



from ctapipe.core import Component
from traitlets import Unicode
from time import sleep

class StringWriter(Component):

    """`StringWriter` class represents a Stage or a Consumer for pipeline.
        It writes received objects to file
    """
    filename = Unicode('/tmp/test.txt', help='output filename').tag(
        config=True, allow_none=True)

    def init(self):
        self.file = open(self.filename, 'w')
        self.log.debug("--- StringWriter init filename {}---".format(self.filename))
        return True

    def run(self, object):
        self.file.write(str(object) + "\n")
        sleep(.5)
        self.log.debug('StringWriter write {}'.format( object))

    def finish(self):
        self.file.close()
        self.log.debug("--- StringWriter finish---")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os
import re
from pkg_resources import resource_listdir
from pathlib import Path
from astropy.utils.decorators import deprecated
import logging

logger = logging.getLogger(__name__)

try:
    import ctapipe_resources
except:
    raise RuntimeError("Please install the 'ctapipe-extra' package, "
                       "which contains the ctapipe_resources module "
                       "needed by ctapipe. (conda install ctapipe-extra)")


__all__ = ['get_dataset', 'find_in_path', 'find_all_matching_datasets']


def get_searchpath_dirs(searchpath=os.getenv("CTAPIPE_SVC_PATH")):
    """ returns a list of dirs in specified searchpath"""
    if searchpath == "" or searchpath is None:
        return []
    return os.path.expandvars(searchpath).split(':')

def find_all_matching_datasets(pattern,
                               searchpath=None,
                               regexp_group=None):
    """
    Returns a list of resource names (or substrings) matching the given 
    pattern, searching first in searchpath (a colon-separated list of 
    directories) and then in the ctapipe_resources module)
    
    Parameters
    ----------
    pattern: str
       regular expression to use for matching
    searchpath: str
       colon-seprated list of directories in which to search, defaulting to 
       CTAPIPE_SVC_PATH environment variable
    regexp_group: int
       if not None, return the regular expression group indicated (assuming 
       pattern has a group specifier in it)

    Returns
    -------
    list(str):
       resources names, use get_dataset() to retrieve the full filename
    """
    results = set()

    if searchpath is None:
        searchpath = os.getenv("CTAPIPE_SVC_PATH")

    # first check search path
    if searchpath is not None:
        for path in get_searchpath_dirs(searchpath):
            if os.path.exists(path):
                for filename in os.listdir(path):
                    match = re.match(pattern, filename)
                    if match:
                        if regexp_group is not None:
                            results.add(match.group(regexp_group))
                        else:
                            results.add(filename)

    # then check resources module
    for resource in resource_listdir('ctapipe_resources', ''):
        match = re.match(pattern, resource)
        if match:
            if regexp_group is not None:
                results.add(match.group(regexp_group))
            else:
                results.add(resource)

    return list(results)


def find_in_path(filename, searchpath):
    """
    Search in searchpath for filename, returning full path.
    
    Parameters
    ----------
    searchpath: str
        colon-separated list of directories (like PATH)
    filename:
        filename to find

    Returns
    -------
    full path to file if found, None otherwise

    """

    for dir in get_searchpath_dirs(searchpath):
        pathname = os.path.join(dir, filename)
        if os.path.exists(pathname):
            return pathname

    return None


def get_dataset(filename):
    """
    Returns the full file path to an auxiliary dataset needed by 
    ctapipe, given the dataset's full name (filename with no directory).
      
    This will first search for the file in directories listed in 
    tne environment variable CTAPIPE_SVC_PATH (if set), and if not found,  
    will look in the ctapipe_resources module 
    (installed with the ctapipe-extra package), which contains the defaults.
    
    Parameters
    ----------
    filename: str
        name of dataset to fetch

    Returns
    -------
    string with full path to the given dataset
    """
    searchpath = os.getenv("CTAPIPE_SVC_PATH")

    if searchpath:
        filepath = find_in_path(filename=filename, searchpath=searchpath)
        if filepath:
            return filepath

    logger.debug("Resource '{}' not found in CTAPIPE_SVC_PATH, looking in "
                 "ctapipe_resources...".format(filename))

    return ctapipe_resources.get(filename)

@deprecated("ctapipe-0.5",alternative='get_dataset()')
def get_path(filename):
    return get_dataset(filename)




import numpy as np
import astropy.units as u
from ctapipe.image.muon.ring_fitter import RingFitter
from ctapipe.io.containers import MuonRingParameter

__all__ =['ChaudhuriKunduRingFitter']

class ChaudhuriKunduRingFitter(RingFitter):

    @u.quantity_input
    def fit(self, x: u.deg, y: u.deg, weight, times=None):
        """Fast and reliable analytical circle fitting method previously used
        in the H.E.S.S.  experiment for muon identification

        Implementation based on [chaudhuri93]_

        Parameters
        ----------
        x: ndarray 
            X position of pixel
        y: ndarray
            Y position of pixel
        weight: ndarray
            weighting of pixel in fit 

        Returns
        -------
        X position, Y position and radius of circle
        """
        # First calculate the weighted average positions of the pixels
        sum_weight = np.sum(weight)
        av_weighted_pos_x = np.sum(x * weight) / sum_weight
        av_weighted_pos_y = np.sum(y * weight) / sum_weight

        # The following notation is a bit ugly but directly references the paper notation
        factor = x**2 + y**2

        a = np.sum(weight * (x - av_weighted_pos_x) * x)
        a_prime = np.sum(weight * (y - av_weighted_pos_y) * x)

        b = np.sum(weight * (x - av_weighted_pos_x) * y)
        b_prime = np.sum(weight * (y - av_weighted_pos_y) * y)

        c = np.sum(weight * (x - av_weighted_pos_x) * factor) * 0.5
        c_prime = np.sum(weight * (y - av_weighted_pos_y) * factor) * 0.5

        nom_0 = ((a * b_prime) - (a_prime * b))
        nom_1 = ((a_prime * b) - (a * b_prime))

        # Calculate circle centre and radius
        centre_x = ((b_prime * c) - (b * c_prime)) / nom_0
        centre_y = ((a_prime * c) - (a * c_prime)) / nom_1

        radius = np.sqrt(
            #np.sum(weight * ((x - centre_x*u.deg)**2 + (y - centre_y*u.deg)**2)) / # centre * u.deg ???
            np.sum(weight * ((x - centre_x)**2 + (y - centre_y)**2)) /
            sum_weight
        )

        output = MuonRingParameter()
        output.ring_center_x = centre_x#*u.deg
        output.ring_center_y = centre_y#*u.deg
        output.ring_radius = radius#*u.deg
        #output.meta.ring_fit_method = "ChaudhuriKundu"
        output.ring_fit_method = "ChaudhuriKundu"

        return output



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Utilities to generate toymodel (fake) reconstruction inputs for testing 
purposes.

Example:

.. code-block:: python

    >>> from instrument import CameraGeometry
    >>> geom = CameraGeometry.make_rectangular(20,20)
    >>> showermodel = generate_2d_shower_model(centroid=[0.25, 0.0], 
    length=0.1,width=0.02, psi='40d')
    >>> image, signal, noise = make_toymodel_shower_image(geom, showermodel.pdf)
    >>> print(image.shape)
    (400,)
                                             
.. plot:: image/image_example.py
    :include-source:



"""
import numpy as np
from scipy.stats import multivariate_normal
from ctapipe.utils import linalg

__all__ = [
    'generate_2d_shower_model',
    'make_toymodel_shower_image',
]


def generate_2d_shower_model(centroid, width, length, psi):
    """Create a statistical model (2D gaussian) for a shower image in a
    camera. The model's PDF (`model.pdf`) can be passed to
    `make_toymodel_shower_image`.

    Parameters
    ----------
    centroid : (float,float)
        position of the centroid of the shower in camera coordinates
    width : float
        width of shower (minor axis)
    length : float
        length of shower (major axis)
    psi : convertable to `astropy.coordinates.Angle`
        rotation angle about the centroid (0=x-axis)

    Returns
    -------

    a `scipy.stats` object

    """
    aligned_covariance = np.array([[length, 0], [0, width]])
    # rotate by psi angle: C' = R C R+
    rotation = linalg.rotation_matrix_2d(psi)
    rotated_covariance = rotation.dot(aligned_covariance).dot(rotation.T)
    return multivariate_normal(mean=centroid, cov=rotated_covariance)


def make_toymodel_shower_image(geom, showerpdf, intensity=50, nsb_level_pe=50):
    """Generates a pedestal-subtracted shower image from a statistical
    shower model (as generated by `shower_model`). The resulting image
    will be in the same format as the given
    `~ctapipe.image.camera.CameraGeometry`.

    Parameters
    ----------
    geom : `ctapipe.instrument.CameraGeometry`
        camera geometry object 
    showerpdf : func
        PDF function for the shower to generate in the camera, e.g. from a 
    intensity : int
        factor to multiply the model by to get photo-electrons
    nsb_level_pe : type
        level of NSB/pedestal in photo-electrons

    Returns
    -------

    an array of image intensities corresponding to the given `CameraGeometry`

    """
    pos = np.empty(geom.pix_x.shape + (2,))
    pos[..., 0] = geom.pix_x.value
    pos[..., 1] = geom.pix_y.value

    model_counts = (showerpdf(pos) * intensity).astype(np.int32)
    signal = np.random.poisson(model_counts)
    noise = np.random.poisson(nsb_level_pe, size=signal.shape)
    image = (signal + noise) - np.mean(noise)

    return image, signal, noise

def Gauss(x,mean,sigma):
    return np.exp(-(x-mean)**2./(2.*sigma*sigma))

def generate_muon_model(xy,radius,width,centre_x,centre_y):
    r_pix = np.sqrt((xy[...,0]-centre_x)**2. + (xy[...,1]-centre_y)**2.)
    Im_pix = Gauss(r_pix,radius,width)
    return Im_pix



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .HillasReconstructor import HillasReconstructor, Reconstructor
from .ImPACT import ImPACTReconstructor
from .energy_regressor import EnergyRegressor
from .shower_max import ShowerMaxEstimator


__all__ = ['HillasReconstructor','Reconstructor','ImPACTReconstructor',
           'EnergyRegressor','ShowerMaxEstimator']


"""
Get version identification from git.


The update_release_version() function writes the current version to the
VERSION file. This function should be called before packaging a release version.

Use the get_version() function to get the version string, including the latest
commit,  from git.
If git is not available the VERSION file will be read.

Heres an example of such a version string:

    v0.2.0.post58+git57440dc


This code was taken from here:
https://github.com/aebrahim/python-git-version

Combining ideas from
http://blogs.nopcode.org/brainstorm/2013/05/20/pragmatic-python-versioning-via-setuptools-and-git-tags/
and Python Versioneer
https://github.com/warner/python-versioneer
but being much more lightwheight

"""
from subprocess import check_output, CalledProcessError
from os import path, name, devnull, environ, listdir

__all__ = ("get_version",)

CURRENT_DIRECTORY = path.dirname(path.abspath(__file__))
VERSION_FILE = path.join(CURRENT_DIRECTORY, "_version_cache.py")

GIT_COMMAND = "git"

if name == "nt":
    def find_git_on_windows():
        """find the path to the git executable on windows"""
        # first see if git is in the path
        try:
            check_output(["where", "/Q", "git"])
            # if this command succeeded, git is in the path
            return "git"
        # catch the exception thrown if git was not found
        except CalledProcessError:
            pass
        # There are several locations git.exe may be hiding
        possible_locations = []
        # look in program files for msysgit
        if "PROGRAMFILES(X86)" in environ:
            possible_locations.append("%s/Git/cmd/git.exe" %
                                      environ["PROGRAMFILES(X86)"])
        if "PROGRAMFILES" in environ:
            possible_locations.append("%s/Git/cmd/git.exe" %
                                      environ["PROGRAMFILES"])
        # look for the github version of git
        if "LOCALAPPDATA" in environ:
            github_dir = "%s/GitHub" % environ["LOCALAPPDATA"]
            if path.isdir(github_dir):
                for subdir in listdir(github_dir):
                    if not subdir.startswith("PortableGit"):
                        continue
                    possible_locations.append("%s/%s/bin/git.exe" %
                                              (github_dir, subdir))
        for possible_location in possible_locations:
            if path.isfile(possible_location):
                return possible_location
        # git was not found
        return "git"

    GIT_COMMAND = find_git_on_windows()


def get_git_describe_version(abbrev=7):
    """return the string output of git desribe"""
    try:
        with open(devnull, "w") as fnull:
            arguments = [GIT_COMMAND, "describe", "--tags",
                         "--abbrev=%d" % abbrev]
            return check_output(arguments, cwd=CURRENT_DIRECTORY,
                                stderr=fnull).decode("ascii").strip()
    except (OSError, CalledProcessError):
        return None


def format_git_describe(git_str, pep440=False):
    """format the result of calling 'git describe' as a python version"""

    if "-" not in git_str:  # currently at a tag
        formatted_str = git_str
    else:
        # formatted as version-N-githash
        # want to convert to version.postN-githash
        git_str = git_str.replace("-", ".post", 1)
        if pep440:  # does not allow git hash afterwards
            formatted_str = git_str.split("-")[0]
        else:
            formatted_str = git_str.replace("-g", "+git")

    # need to remove the "v" to have a proper python version
    if formatted_str.startswith('v'):
        formatted_str = formatted_str[1:]

    return formatted_str


def read_release_version():
    """Read version information from VERSION file"""
    try:
        from ._version_cache import version
        if len(version) == 0:
            version = None
        return version
    except ImportError:
        return "unknown"


def update_release_version(pep440=False):
    """Release versions are stored in a file called VERSION.
    This method updates the version stored in the file.
    This function should be called when creating new releases.
    It is called by setup.py when building a package.


    pep440: bool
        When True, this function returns a version string suitable for
        a release as defined by PEP 440. When False, the githash (if
        available) will be appended to the version string.

    """
    version = get_version(pep440=pep440)
    with open(VERSION_FILE, "w") as outfile:
        outfile.write("version='{}'".format(version))
        outfile.write("\n")


def get_version(pep440=False):
    """Tracks the version number.

    pep440: bool
        When True, this function returns a version string suitable for
        a release as defined by PEP 440. When False, the githash (if
        available) will be appended to the version string.

    The file VERSION holds the version information. If this is not a git
    repository, then it is reasonable to assume that the version is not
    being incremented and the version returned will be the release version as
    read from the file.

    However, if the script is located within an active git repository,
    git-describe is used to get the version information.

    The file VERSION will need to be changed manually.
    """

    raw_git_version = get_git_describe_version()
    if not raw_git_version:  # not a git repository
        return  read_release_version()

    git_version = format_git_describe(raw_git_version, pep440=pep440)

    return git_version


if __name__ == "__main__":
    print(get_version())



"""
Plot charge resolution data generated by ctapipe-chargeres-extract.
"""

from os.path import dirname, basename, expanduser, exists, splitext
from os import makedirs
from matplotlib import pyplot as plt
from math import log10
import warnings
import numpy as np
from traitlets import Dict, List, Unicode, Int, Bool
from ctapipe.core import Tool, Component
from ctapipe.analysis.camera.chargeresolution import ChargeResolutionCalculator


class ChargeResolutionPlotter(Component):
    name = 'ChargeResolutionPlotter'

    output_path = Unicode(None, allow_none=True,
                          help='Output path to save the '
                               'plot.').tag(config=True)
    max_pe = Int(2000, help='Maximum pe to plot').tag(config=True)
    linear_x = Bool(False, help='Plot the x values on a linear axis, '
                                'instead of log').tag(config=True)
    linear_y = Bool(False, help='Plot the y values on a linear axis, '
                                'instead of log').tag(config=True)

    def __init__(self, config, tool, **kwargs):
        """
        Calculator of charge resolution.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        reductor : ctapipe.calib.camera.reductors.Reductor
            The reductor to use to reduce the waveforms in the event.
            By default no data volume reduction is applied, and the dl0 samples
            will equal the r1 samples.
        kwargs
        """
        super().__init__(config=config, parent=tool, **kwargs)

        try:
            if self.output_path is None:
                raise ValueError
        except ValueError:
            self.log.exception('Please specify an output path')
            raise

        self.fig = plt.figure(figsize=(20, 8))
        self.ax_l = self.fig.add_subplot(121)
        self.ax_r = self.fig.add_subplot(122)

        self.fig.subplots_adjust(left=0.05, right=0.95, wspace=0.6)

        self.legend_handles = []
        self.legend_labels = []

    def plot_chargeres(self, name, x, res, error):
        valid = (x <= self.max_pe)
        x_v = x[valid]
        res_v = res[valid]
        error_v = error[valid]

        eb_l, _, _ = self.ax_l.errorbar(x_v, res_v, yerr=error_v,
                                        marker='x', linestyle="None")
        self.legend_handles.append(eb_l)
        self.legend_labels.append(splitext(name)[0])

    def plot_scaled_chargeres(self, x, res, error):
        valid = (x <= self.max_pe)
        x_v = x[valid]
        res_v = res[valid]
        error_v = error[valid]

        self.ax_r.errorbar(x_v, res_v, yerr=error_v,
                           marker='x', linestyle="None")

    def plot_limit_curves(self):
        x = np.logspace(log10(0.9), log10(self.max_pe * 1.1), 100)
        requirement = ChargeResolutionCalculator.requirement(x)
        goal = ChargeResolutionCalculator.goal(x)
        poisson = ChargeResolutionCalculator.poisson(x)

        r_p, = self.ax_l.plot(x, requirement, 'r', ls='--')
        g_p, = self.ax_l.plot(x, goal, 'g', ls='--')
        p_p, = self.ax_l.plot(x, poisson, c='0.75', ls='--')
        self.ax_r.plot(x, requirement / goal, 'r')
        self.ax_r.plot(x, np.ones_like(x), 'g')
        self.ax_r.plot(x, poisson / goal, c='0.75', ls='--')

        self.legend_handles.append(r_p)
        self.legend_labels.append("Requirement")
        self.legend_handles.append(g_p)
        self.legend_labels.append("Goal")
        self.legend_handles.append(p_p)
        self.legend_labels.append("Poisson")

    def save(self):
        if not self.linear_x:
            self.ax_l.set_xscale('log')
            self.ax_r.set_xscale('log')
        if not self.linear_y:
            self.ax_l.set_yscale('log')

        self.ax_l.set_xlabel(r'True Charge $Q_T$ (p.e.)')
        self.ax_l.set_ylabel('Charge Resolution')
        self.ax_r.set_xlabel(r'True Charge $Q_T$ (p.e.)')
        self.ax_r.set_ylabel('Charge Resolution/Goal')

        self.ax_l.legend(self.legend_handles, self.legend_labels,
                         bbox_to_anchor=(1.02, 1.), loc=2,
                         borderaxespad=0., fontsize=10)

        self.fig.suptitle(splitext(basename(self.output_path))[0])

        self.ax_l.set_xlim(0.9, self.max_pe * 1.1)
        self.ax_r.set_xlim(0.9, self.max_pe * 1.1)
        if self.max_pe > 2000:
            self.ax_r.set_xlim(0.9, 2000 * 1.1)
        # if args.maxpeplot is not None:
        #     ax_l.set_xlim(0.9, args.maxpeplot)
        #     ax_r.set_xlim(0.9, args.maxpeplot)

        warnings.filterwarnings("ignore", module="matplotlib")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if self.output_path is None:
                plt.show()
            else:
                self.output_path = expanduser(self.output_path)
                output_dir = dirname(self.output_path)
                if not exists(output_dir):
                    self.log.info("[output] Creating directory: "
                                  "{}".format(output_dir))
                    makedirs(output_dir)
                self.log.info("[output] {}".format(self.output_path))
                self.fig.savefig(self.output_path, bbox_inches='tight')


class ChargeResolutionViewer(Tool):
    name = "ChargeResolutionViewer"
    description = "Plot the charge resolution from " \
                  "ChargeResolutionCalculator objects restored via " \
                  "pickled dictionaries."

    input_files = List(Unicode, None,
                       help='Input pickle files that are produced from '
                            'ChargeResolutionCalculator.save().'
                            '').tag(config=True)

    aliases = Dict(dict(f='ChargeResolutionViewer.input_files',
                        B='ChargeResolutionCalculator.binning',
                        max_pe='ChargeResolutionPlotter.max_pe',
                        O='ChargeResolutionPlotter.output_path',
                        ))
    flags = Dict(dict(L=({'ChargeResolutionCalculator': {'log_bins': False}},
                         'Bin the x axis linearly instead of logarithmic.'),
                      linx=({'ChargeResolutionPlotter': {'linear_x': True}},
                            'Plot the x values on a linear axis, '
                            'instead of log.'),
                      liny=({'ChargeResolutionPlotter': {'linear_y': True}},
                            'Plot the x values on a linear axis, '
                            'instead of log.')
                      ))
    classes = List([ChargeResolutionCalculator,
                    ChargeResolutionPlotter
                    ])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.calculator = None
        self.plotter = None

    def setup(self):
        self.log_format = "%(levelname)s: %(message)s [%(name)s.%(funcName)s]"
        kwargs = dict(config=self.config, tool=self)

        self.calculator = ChargeResolutionCalculator(**kwargs)
        self.plotter = ChargeResolutionPlotter(**kwargs)

    def start(self):
        self.plotter.plot_limit_curves()
        for fp in self.input_files:
            self.calculator.load(fp)
            x, res, res_error, scaled_res, scaled_res_error = \
                self.calculator.get_charge_resolution()

            name = basename(fp)
            self.plotter.plot_chargeres(name, x, res, res_error)
            self.plotter.plot_scaled_chargeres(x, scaled_res, scaled_res_error)

    def finish(self):
        self.plotter.save()


def main():
    exe = ChargeResolutionViewer()
    exe.run()

if __name__ == '__main__':
    main()



import json

from ctapipe.core import Provenance
from ctapipe.core.provenance import _ActivityProvenance


def test_Provenance():
    prov = Provenance()
    prov.start_activity("test1")
    prov.add_input_file("input.txt")
    prov.add_output_file("output.txt")
    prov.start_activity("test2")
    prov.add_input_file("input_a.txt")
    prov.add_input_file("input_b.txt")
    prov.finish_activity("test2")
    prov.finish_activity("test1")

    assert set(prov.finished_activity_names) == {'test2', 'test1'}

    return prov


def test_ActivityProvenance():
    prov = _ActivityProvenance()
    prov.start()
    prov.register_input('test.txt')
    prov.register_input('test2.txt')
    prov.register_output('out.txt')
    prov.sample_cpu_and_memory()
    prov.finish()


def test_provenence_contextmanager():

    prov = Provenance()

    with prov.activity("myactivity"):
        assert 'myactivity' in prov.active_activity_names

    assert 'myactivity' in prov.finished_activity_names
    assert 'myactivity' not in prov.active_activity_names


if __name__ == '__main__':

    import logging
    logging.basicConfig(level=logging.DEBUG)

    prov = test_Provenance()
    print(json.dumps(prov.provenance, indent=4))



"""
Serialize ctapipe containers to file
"""

from abc import ABC, abstractmethod
from gzip import open as gzip_open
from pickle import dump

import numpy as np
from astropy import log
from astropy.table import Table, Column
from traitlets import Unicode

from ctapipe.core import Container

__all__ = ['Serializer']


class Serializer:
    """
    Serializes ctapipe.core.Component, write it to a file thanks
    to its Writer object
    For some formats (i.e. pickle +gzip), read serialized components from
    a file

    Examples
    --------
    >>> writer = Serializer(filename='output.pickle', format='pickle', mode='w')
    >>> for container in input_containers:
    ...       writer.add_container(container.r0)
    >>> writer.close()

    or using the context manager syntax
    >>> with Serializer(filename='output.fits', format='fits', mode='w') as writer:
    >>> for container in input_containers:
    ...     writer.add_container(container.r0)
    """

    def __init__(self, filename, format='fits', mode='x'):

        """
        Parameters
        ----------
        filename: str
            full path name for i/o file
        format: str ('fits', 'img', 'pickle')
        mode: str ('write', 'read')
            : use this serializer as writer or reader
        mode: str
            'w'	open for writing, truncating the file first
            'x'	open for exclusive creation, failing if the file already exists
            'a'	open for writing, appending to the end of the file if it exists
        Raises
        ------
        NotImplementedError: when format is not implemented
        ValueError: when mode is not correct
        """
        self.filename = filename
        self.format = format
        self._stat = None # TODO collect statistics about serialized contents
        if mode not in ('x', 'w', 'a'):
            raise ValueError('{} is not a valid write mode. Use x, w or a'.
                             format(mode))
        self._writer = None

        if self.format == 'fits':
            self._writer = TableWriter(outfile=filename, mode=mode,
                                       format=format)
        elif self.format == 'pickle':
            self._writer = GZipPickleWriter(outfile=filename, mode=mode)
        elif self.format == 'img':
            raise NotImplementedError('img serializer format is'
                                      ' not yet implemented')
        else:
            raise ValueError('You can serialize only on pickle, fits or img')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the runtime context related to this object.
        The parameters describe the exception that caused the context to be
        exited. If the context was exited without an exception,
        all three arguments will be None.
        If an exception is supplied, and the method wishes to suppress
        the exception (i.e., prevent it from being propagated),
        it should return a true value. Otherwise, the exception will be
        processed normally upon exit from this method.
        """
        self.close()

    def add_container(self, container):
        """
        Add a container to serializer
        """
        self._writer.add_container(container)


    def close(self):
        """
        Write data to disk
        """
        self._writer.close()


class Writer(ABC):
    def __init__(self, filename):
        self.outfile = filename

    @abstractmethod
    def add_container(self, container):
        pass

    @abstractmethod
    def close(self):
        pass


class GZipPickleWriter(Writer):
    """
    Serializes list of ctapipe.core.Components.
    Write Component to file
    """

    def __init__(self, outfile, mode='x'):
        """
        Parameters
        ----------
        outfile:  Unicode
            full path output file name
        mode: str
            'w'	open for writing, truncating the file first
            'x'	open for exclusive creation, failing if the file already exists
            'a'	open for writing, appending to the end of the file if it exists
        Raises
        ------
        FileNotFoundError: When the file cannot be opened
        FileExistsError: when infile exist and mode is x
        """
        super().__init__(outfile)
        mode += 'b'
        try:
            self.file_object = gzip_open(outfile, mode)
        except FileExistsError:
            raise FileExistsError('file exists: {} and mode is {}'.
                                  format(outfile, mode))

    def close(self):
        """
        close opened file
        Returns
        -------
        """
        self.file_object.close()

    def add_container(self, container):
        """
        Add a container to be serialized

        Raises
        ------
        TypeError: When container is not type of container
        """
        if not isinstance(container, Container):
            raise TypeError('Can write only Containers')
        dump(container, self.file_object)


# FITS Implementation

not_writeable_fields = ('tel', 'tels_with_data', 'calibration_parameters',
                        'pedestal_subtracted_adc', 'integration_window')


def is_writeable(key, out_format='fits'):
    """
    check if a key is writable

    Parameters
    ----------
    key: str
    out_format: 'fits' or ´ pickle'
        according to out_format a same key can be writable or not

    Returns
    -------
    True if key is writable according to the out_format

    Raises
    ------
    NameError: When out_format is not know
    """
    if out_format is 'fits':
        return not (key in not_writeable_fields)
    elif out_format is 'pickle':
        return True
    else:
        raise NameError('{} not implemented'.format(out_format))


def writeable_items(container):
    """
    # Strip off what we cannot write
    Parameters
    ----------
    container: ctapipe.core.Container

    Returns
    -------
    a dictionary with writable values only
    """

    d = dict(container.items())
    for k in not_writeable_fields:
        log.debug("Cannot write column {0}".format(k))
        d.pop(k, None)
    return d


def to_table(container):
    """
    Convert a `ctapipe.core.Container` to an `astropy.Table` with one row

    Parameters
    ----------
    container: ctapipe.core.Container

    Returns
    -------
    Table: astropy.Table
    """
    names = list()
    columns = list()
    for k, v in writeable_items(container).items():

        v_arr = np.array(v)
        v_arr = v_arr.reshape((1,) + v_arr.shape)
        log.debug("Creating column for item '{0}' of shape {1}".
                  format(k, v_arr.shape))
        names.append(k)
        columns.append(Column(v_arr))

    return Table(data=columns,  # dtypes are inferred by columns
                 names=names,
                 meta=container.meta)


class TableWriter(Writer):
    """
    Fits table writer
    """
    def __init__(self, outfile, format='fits', mode='w'):
        """
        Parameters
        ----------
        outfile: str
            output file name
        format: str
            'fits' or 'img'
        mode: str
            'w'	open for writing, truncating the file first
            'x'	open for exclusive creation, failing if the file already exists
        Raises
        ------
        NotImplementedError: when mode is correct but not yet implemented
        ValueError: when mode is not correct
        """
        super().__init__(outfile)
        self.table = Table()
        self._created_table = False
        self.format = format
        self.outfile = outfile
        if mode == 'w':
            self.overwrite = True
        elif mode == 'x':
            self.overwrite = False
        elif mode == 'a':
            raise NotImplementedError('a is a valid write mode,'
                                      ' but not yet implemented')
        else:
            raise ValueError('{} is not a valid write mode. Use x, w or a'.
                             format(mode))

    def _setup_table(self, container):
        """
        Create Fits table and HDU

        Parameters
        ----------
        container: ctapipe.core.Container
        """
        # Create Table from Container
        self.table = to_table(container)

        # Write HDU name
        if self.format == "fits":
            self.table.meta["EXTNAME"] = type(container).__name__
        self._created_table = True

    def add_container(self, container):
        """
        Add a container as a table row
        Parameters
        ----------
        container: ctapipe.core.Container

        Raises
        ------
        TypeError: When add another type than Container
        """
        if not isinstance(container, Container):
            raise TypeError("Can write only Containers")

        if not self._created_table:
            self._setup_table(container)
        else:
            self.table.add_row(writeable_items(container))

    def close(self, **kwargs):
        """
        Write Fits table to file
        Parameters
        ----------
        kwargs to be passed to `astropy.Table.write method`

        Returns
        -------
        Fits Table
        """
        # Write table using astropy.table write method
        self.table.write(output=self.outfile, format=self.format,
                         overwrite=self.overwrite, **kwargs)
        return self.table



class Singleton(type):
    """" metaclass for singleton pattern """
    instance = None

    def __call__(cls, *args, **kw):
        if not cls.instance:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance



import numpy as np
from ctapipe.reco.ImPACT import ImPACTReconstructor
from ctapipe.utils import get_dataset
import astropy.units as u
from numpy.testing import assert_allclose
from ctapipe.io.containers import ReconstructedShowerContainer, ReconstructedEnergyContainer
import pytest

class TestImPACT():

    @classmethod
    def setup_class(self):
        self.impact_reco = ImPACTReconstructor(root_dir=".")

    @pytest.mark.skip('need a dataset for this to work')
    def test_brightest_mean_average(self):
        """
        Test that averaging of the brightest pixel position give a sensible outcome
        """
        image = np.array([1,1,1,1])
        pixel_x = np.array([0.,1.,0.,-1.])*u.deg
        pixel_y = np.array([-1.,0.,1.,0.])*u.deg
        pixel_area = np.array([0,0,0,0])*u.deg*u.deg

        self.impact_reco.set_event_properties({1:image}, {1:pixel_x},
                                              {1:pixel_y}, {1:pixel_area},
                                              {1:"CHEC"}, {1:0*u.m},
                                              {1:0*u.m}, {1:0*u.m})

        self.impact_reco.get_brightest_mean(num_pix=pixel_x.shape[0])

        assert_allclose(self.impact_reco.peak_x[0], 0, rtol=0, atol=0.001)
        assert_allclose(self.impact_reco.peak_y[0], 0, rtol=0, atol=0.001)

    @pytest.mark.skip('need a dataset for this to work')
    def test_brightest_mean_sort(self):
        """
        Test that pixels are sorted into the correct order
        """
        image = np.array([1,1,1,0])
        pixel_x = np.array([1,1,1,0])*u.rad
        pixel_y = np.array([1,1,1,0.])*u.rad
        pixel_area = np.array([0,0,0,0])*u.deg*u.deg

        self.impact_reco.set_event_properties({1:image}, {1:pixel_x}, {1:pixel_y}, {1:pixel_area},
                                              {1:"CHEC"}, {1:0*u.m}, {1:0*u.m}, {1:0*u.m})
        self.impact_reco.get_brightest_mean(num_pix=3)

        assert_allclose(self.impact_reco.peak_x[0], 1, rtol=0, atol=0.001)
        assert_allclose(self.impact_reco.peak_y[0], 1, rtol=0, atol=0.001)

    def test_rotation(self):
        """Test pixel rotation function"""
        x = np.array([1])
        y = np.array([0])

        xt, yt = ImPACTReconstructor.rotate_translate(x, y, 0, 0, np.deg2rad(90))
        assert_allclose(xt, 0, rtol=0, atol=0.001)
        assert_allclose(yt, 1, rtol=0, atol=0.001)

        xt, yt = ImPACTReconstructor.rotate_translate(x, y, 0, 0, np.deg2rad(180))
        assert_allclose(xt, -1, rtol=0, atol=0.001)
        assert_allclose(yt, 0, rtol=0, atol=0.001)

    def test_translation(self):
        """Test pixel translation function"""
        x = np.array([0])
        y = np.array([0])

        xt, yt = ImPACTReconstructor.rotate_translate(x, y, 1, 1, 0)
        assert_allclose(xt, -1, rtol=0, atol=0.001)
        assert_allclose(yt, -1, rtol=0, atol=0.001)

    @pytest.mark.skip('need a dataset for this to work')
    def test_xmax_calculation(self):
        """Test calculation of hmax and interpolation of Xmax tables"""

        image = np.array([1,1,1])
        pixel_x = np.array([1,1,1])*u.deg
        pixel_y = np.array([1,1,1])*u.deg
        pixel_area = np.array([0,0,0])*u.deg*u.deg

        self.impact_reco.set_event_properties({1:image}, {1:pixel_x},
                                              {1:pixel_y}, {1:pixel_area},
                                              {1:"CHEC"}, {1:0*u.m},
                                              {1:0*u.m},
                                              array_direction=[0*u.deg,0*u.deg])

        max = self.impact_reco.get_shower_max(0, 0, 0, 100, 0)

        assert_allclose(max, 486.85820802*u.g/(u.cm*u.cm), rtol=0.01)

    @pytest.mark.skip('need a dataset for this to work')
    def test_image_prediction(self):

        pixel_x = np.array([0])*u.deg
        pixel_y = np.array([0])*u.deg

        image = np.array([1])
        pixel_area = np.array([1])*u.deg*u.deg

        self.impact_reco.set_event_properties({1:image}, {1:pixel_x},
                                              {1:pixel_y}, {1:pixel_area},
                                              {1:"CHEC"}, {1:0*u.m}, {1:0*u.m},
                                              array_direction=[0*u.deg,0*u.deg])

        """First check image prediction by directly accessing the function"""
        pred = self.impact_reco.image_prediction("CHEC", zenith=0, azimuth=0,
                                                 energy=1, impact=50, x_max=0,
                                                 pix_x=pixel_x, pix_y=pixel_y)

        assert np.sum(pred) != 0

        """Then check helper function gives the same answer"""
        shower = ReconstructedShowerContainer()
        shower.is_valid=True
        shower.alt = 0*u.deg
        shower.az = 0*u.deg
        shower.core_x = 0 *u.m
        shower.core_y = 100*u.m
        shower.h_max = 300 + 93 * np.log10(1)

        energy = ReconstructedEnergyContainer()
        energy.is_valid = True
        energy.energy = 1*u.TeV
        pred2 = self.impact_reco.get_prediction(1, shower_reco=shower,
                                                energy_reco=energy)
        print(pred, pred2)
        assert pred.all() == pred2.all()

    @pytest.mark.skip('need a dataset for this to work')
    def test_likelihood(self):

        pixel_x = np.array([0])*u.deg
        pixel_y = np.array([0])*u.deg

        image = np.array([1])
        pixel_area = np.array([1])*u.deg*u.deg

        self.impact_reco.set_event_properties({1:image}, {1:pixel_x},
                                              {1:pixel_y}, {1:pixel_area},
                                              {1:"CHEC"}, {1:0*u.m}, {1:0*u.m},
                                              array_direction=[0*u.deg,0*u.deg])

        like = self.impact_reco.get_likelihood(0, 0, 0, 100, 1, 0)
        assert like is not np.nan and like > 0


"""
Image Cleaning Algorithms (identification of noisy pixels)
"""

__all__ = ['tailcuts_clean', 'dilate']

import numpy as np

def tailcuts_clean(geom, image, picture_thresh=7, boundary_thresh=5,
                   keep_isolated_pixels=False):
    """Clean an image by selection pixels that pass a two-threshold
    tail-cuts procedure.  The picture and boundary thresholds are
    defined with respect to the pedestal dispersion. All pixels that
    have a signal higher than the picture threshold will be retained,
    along with all those above the boundary threshold that are
    neighbors of a picture pixel.

    To include extra neighbor rows of pixels beyond what are accepted, use the
    `ctapipe.reco.dialate` function.

    Parameters
    ----------
    geom: `ctapipe.instrument.CameraGeometry`
        Camera geometry information
    image: array
        pixel values
    picture_thresh: float or array
        threshold above which all pixels are retained
    boundary_thresh: float or array
        threshold above which pixels are retained if they have a neighbor 
        already above the picture_thresh
    keep_isolated_pixels: bool
        If True, pixels above the picture threshold will be included always, 
        if not they are only included if a neighbor is in the picture or 
        boundary

    Returns
    -------

    A boolean mask of *clean* pixels.  To get a zero-suppressed image and pixel
    list, use `image[mask], geom.pix_id[mask]`, or to keep the same
    image size and just set unclean pixels to 0 or similar, use
    `image[~mask] = 0`

    """

    pixels_in_picture = image >= picture_thresh

    # by broadcasting together pixels_in_picture (1d) with the neighbor
    # matrix (2d), we find all pixels that are above the boundary threshold
    # AND have any neighbor that is in the picture
    pixels_above_boundary = image >= boundary_thresh
    pixels_with_picture_neighbors = (pixels_in_picture &
                                     geom.neighbor_matrix).any(axis=1)

    if keep_isolated_pixels:
        return (pixels_above_boundary
                & pixels_with_picture_neighbors) | pixels_in_picture
    else:
        pixels_with_boundary_neighbors = (pixels_above_boundary &
                                         geom.neighbor_matrix).any(axis=1)
        return ((pixels_above_boundary & pixels_with_picture_neighbors) |
                (pixels_in_picture &  pixels_with_boundary_neighbors))




def dilate(geom, mask):
    """
    Add one row of neighbors to the True values of a pixel mask and return 
    the new mask.
    This can be used to include extra rows of pixels in a mask that was
    pre-computed, e.g. via `tailcuts_clean`.
    
    Parameters
    ----------
    geom: `~ctapipe.instrument.CameraGeometry`
        Camera geometry information
    mask: ndarray 
        input mask (array of booleans) to be dilated
    """
    return mask | (mask & geom.neighbor_matrix).any(axis=1)
